"""MeSH qualifiers from raw/mesh/q2020.bin."""

# 76 items in total:
# 76 SH Name
# 76 QA Abbreviation
# 76 QE Short Form
# 76 MS Scope Note

from collections import namedtuple

NtQbin = namedtuple('NtQbin', 'SH QA QE MS')
NTS = [
  NtQbin._make(['abnormalities', 'AB', 'ABNORM', 'Used with organs for congenital defects producing changes in the morphology of the organ. It is used also for abnormalities in animals.']),
  NtQbin._make(['administration & dosage', 'AD', 'ADMIN', 'Used with drugs for dosage forms, routes of administration, frequency and duration of administration, quantity of medication, and the effects of these factors.']),
  NtQbin._make(['adverse effects', 'AE', 'ADV EFF', 'Used with drugs, chemicals, or biological agents in accepted dosage - or with physical agents or manufactured products in normal usage - when intended for diagnostic, therapeutic, prophylactic, or anesthetic purposes. It is used also for adverse effects or complications of diagnostic, therapeutic, prophylactic, anesthetic, surgical, or other procedures.']),
  NtQbin._make(['agonists', 'AG', 'AGON', 'Used with chemicals, drugs, and endogenous substances to indicate substances or agents that have affinity for a receptor and intrinsic activity at that receptor. (From Textbook of Pharmacology, 1991, p.16)']),
  NtQbin._make(['analogs & derivatives', 'AA', 'ANALOGS', 'Used with drugs and chemicals for substances that share the same parent molecule or have similar electronic structure but differ by the addition or substitution of other atoms or molecules. It is used when the specific chemical heading is not available and no appropriate group heading exists.']),
  NtQbin._make(['analysis', 'AN', 'ANAL', 'Used for the identification or quantitative determination of a substance or its constituents and metabolites; includes the analysis of air, water, or other environmental carrier. It excludes the chemical analysis of tissues, tumors, body fluids, organisms, and plants for which "chemistry" is used. The concept applies to both methodology and results. For analysis of substances in blood, cerebrospinal fluid, and urine the specific subheading designating the fluid is used.']),
  NtQbin._make(['anatomy & histology', 'AH', 'ANAT', 'Used with organs, regions, and tissues for normal descriptive anatomy and histology, and for the normal anatomy and structure of animals and plants.']),
  NtQbin._make(['antagonists & inhibitors', 'AI', 'ANTAG', 'Used with chemicals, drugs, and endogenous substances to indicate substances or agents which counteract their biological effects by any mechanism.']),
  NtQbin._make(['biosynthesis', 'BI', 'BIOSYN', 'Used for the anabolic formation of chemical substances in organisms, in living cells, or by subcellular fractions.']),
  NtQbin._make(['blood', 'BL', 'BLOOD', 'Used for the presence or analysis of substances in the blood; also for examination of, or changes in, the blood in disease states. It excludes serodiagnosis, for which the subheading "diagnosis" is used, and serology, for which "immunology" is used.']),
  NtQbin._make(['blood supply', 'BS', 'BLOOD SUPPLY', 'Used for arterial, capillary, and venous systems of an organ or region whenever the specific heading for the vessel does not exist. It includes blood flow through the organ.']),
  NtQbin._make(['cerebrospinal fluid', 'CF', 'CSF', 'Used for the presence or analysis of substances in the cerebrospinal fluid; also for examination of or changes in cerebrospinal fluid in disease states.']),
  NtQbin._make(['chemical synthesis', 'CS', 'CHEM SYN', 'Used for the chemical preparation of molecules in vitro. For the formation of chemical substances in organisms, living cells, or subcellular fractions, "biosynthesis" is used.']),
  NtQbin._make(['chemically induced', 'CI', 'CHEM IND', 'Used for biological phenomena, diseases, syndromes, congenital abnormalities, or symptoms caused by endogenous or exogenous substances.']),
  NtQbin._make(['chemistry', 'CH', 'CHEM', 'Used with chemicals, biological, and non-biological substances for their composition, structure, characterization, and properties; also used for the chemical composition or content of organs, tissue, tumors, body fluids, organisms, and plants. Excludes chemical analysis and determination of substances for which "analysis" is used; excludes synthesis for which "chemical synthesis" is used; excludes isolation and purification of substances for which "isolation & purification" is used.']),
  NtQbin._make(['classification', 'CL', 'CLASS', 'Used for taxonomic or other systematic or hierarchical classification systems.']),
  NtQbin._make(['complications', 'CO', 'COMPL', 'Used with diseases to indicate conditions that co-exist or follow, i.e., co-existing diseases, complications, or sequelae.']),
  NtQbin._make(['congenital', 'CN', 'CONGEN', 'Used with disease headings to indicate those conditions existing at, and usually before, birth. It excludes morphologic abnormalities and birth injuries, for which "abnormalities" and "injuries" are used.']),
  NtQbin._make(['cytology', 'CY', 'CYTOL', 'Used for cellular appearance of unicellular and multicellular organisms.']),
  NtQbin._make(['deficiency', 'DF', 'DEFIC', 'Used with endogenous and exogenous substances which are absent or in diminished amount relative to the normal requirement of an organism or a biologic system.']),
  NtQbin._make(['diagnosis', 'DI', 'DIAG', 'Used with diseases for all aspects of diagnosis, including examination, differential diagnosis and prognosis. Excludes diagnosis using imaging techniques (e.g. radiography, scintigraphy, and ultrasonography) for which "diagnostic imaging" is used.']),
  NtQbin._make(['diagnostic imaging', 'DG', 'DIAG IMAGE', 'Used for the visualization of an anatomical structure or for the diagnosis of disease.  Commonly used imaging techniques include radiography, radionuclide imaging, thermography, tomography, and ultrasonography']),
  NtQbin._make(['diet therapy', 'DH', 'DIET THER', 'Used with disease headings for dietary and nutritional management of the disease. The concept does not include vitamin or mineral supplements, for which "drug therapy" may be used.']),
  NtQbin._make(['drug effects', 'DE', 'DRUG EFF', 'Used with organs, regions, tissues, or organisms and physiological and psychological processes for the effects of drugs and chemicals.']),
  NtQbin._make(['drug therapy', 'DT', 'DRUG THER', 'Used with disease headings for the treatment of disease by the administration of drugs, chemicals, and antibiotics. For diet therapy and radiotherapy, use  specific subheadings. Excludes immunotherapy for which "therapy" is used.']),
  NtQbin._make(['economics', 'EC', 'ECON', 'Used for the economic aspects of any subject, as well as for all aspects of financial management. It includes the raising or providing of funds.']),
  NtQbin._make(['education', 'ED', 'EDUC', 'Used for education, training programs, and courses in various fields and disciplines, and for training groups of persons.']),
  NtQbin._make(['embryology', 'EM', 'EMBRYOL', 'Used with organs, regions, and animal headings for embryologic and fetal development. It is used also with diseases for embryologic factors contributing to postnatal disorders.']),
  NtQbin._make(['enzymology', 'EN', 'ENZYMOL', 'Used with organisms, except vertebrates, and with organs and tissues. It is also used with diseases for enzymes during the course of the disease, but excludes diagnostic enzyme tests, for which "diagnosis" is used.']),
  NtQbin._make(['epidemiology', 'EP', 'EPIDEMIOL', 'Used with human and veterinary diseases for the distribution of disease, factors which cause disease, and the attributes of disease in defined populations; includes incidence, frequency, prevalence, endemic and epidemic outbreaks; also surveys and estimates of morbidity in geographic areas and in specified populations. Used also with geographical headings for the location of epidemiologic aspects of a disease. Excludes mortality for which "mortality" is used.']),
  NtQbin._make(['ethics', 'ES', 'ETHICS', 'Used with techniques and activities for discussion and analysis with respect to human and social values.']),
  NtQbin._make(['ethnology', 'EH', 'ETHNOL', 'Used with diseases for ethnic, cultural, or anthropological aspects, and with geographic headings to indicate the place of origin of a group of people.']),
  NtQbin._make(['etiology', 'ET', 'ETIOL', 'Used with diseases for causative agents including microorganisms and includes environmental and social factors and personal habits as contributing factors. It includes pathogenesis.']),
  NtQbin._make(['genetics', 'GE', 'GENET', 'Used for mechanisms of heredity and the genetics of organisms, for the genetic basis of normal and pathologic states, and for the genetic aspects of endogenous chemicals. It includes biochemical and molecular influence on genetic material.']),
  NtQbin._make(['growth & development', 'GD', 'GROWTH', 'Used with microorganisms, plants, and the postnatal period of animals for growth and development. It includes also the postnatal growth or development of organs or anatomical parts. For prenatal period of animals for growth and development use /embryology.']),
  NtQbin._make(['history', 'HI', 'HIST', 'Used for the historical aspects of any subject. It includes brief historical notes but excludes case histories.']),
  NtQbin._make(['immunology', 'IM', 'IMMUNOL', 'Used for immunologic studies of tissues, organs, microorganisms, fungi, viruses, and animals. It includes immunologic aspects of diseases but not immunologic procedures used for diagnostic, preventive, or therapeutic purposes, for which "diagnosis", "prevention & control", or "therapy" are used. The concept is also used for chemicals as antigens or haptens.']),
  NtQbin._make(['injuries', 'IN', 'INJ', 'Used with anatomic headings, animals, and sports for wounds and injuries. Excludes cell damage, for which "pathology" is used.']),
  NtQbin._make(['innervation', 'IR', 'INNERV', 'Used with organs, regions, or tissues for their nerve supply.']),
  NtQbin._make(['instrumentation', 'IS', 'INSTRUM', 'Used with diagnostic or therapeutic procedures, analytic techniques, and specialties or disciplines, for the development or modification of apparatus, instruments, or equipment.']),
  NtQbin._make(['isolation & purification', 'IP', 'ISOL', 'Used with bacteria, viruses, fungi, protozoa, and helminths for the obtaining of pure strains or for the demonstration of the presence of or identification of organisms by DNA analyses, immunologic, or other methods, including culture techniques. It is used also with biological substances and chemicals for the isolation and purification of the constituents.']),
  NtQbin._make(['legislation & jurisprudence', 'LJ', 'LEGIS', 'Used for laws, statutes, ordinances, or government regulations, as well as for legal controversy and court decisions.']),
  NtQbin._make(['metabolism', 'ME', 'METAB', 'Used with organs, cells and subcellular fractions, organisms, and diseases for biochemical changes. It is used also with drugs and chemicals for catabolic changes (breakdown of complex molecules into simpler ones). For anabolic processes (conversion of small molecules into large), BIOSYNTHESIS is used. For enzymology and pharmacokinetics use the specific subheadings.']),
  NtQbin._make(['methods', 'MT', 'METHODS', 'Used with techniques, procedures, and programs for methods.']),
  NtQbin._make(['microbiology', 'MI', 'MICROBIOL', 'Used with organs, animals, and higher plants and with diseases for microbiologic studies. For parasites, "parasitology" is used; for viruses, "virology" is used.']),
  NtQbin._make(['mortality', 'MO', 'MORTAL', 'Used with human and veterinary diseases for mortality statistics. For deaths resulting from various procedures statistically but for a death resulting in a specific case, use FATAL OUTCOME, not /mortality.']),
  NtQbin._make(['nursing', 'NU', 'NURS', 'Used with diseases for nursing care and techniques in their management. It includes the nursing role in diagnostic, therapeutic, and preventive procedures.']),
  NtQbin._make(['organization & administration', 'OG', 'ORGAN', 'Used for administrative structure and management.']),
  NtQbin._make(['parasitology', 'PS', 'PARASITOL', 'Used with animals, higher plants, organs, and diseases for parasitic factors. In diseases, it is not used if the parasitic involvement is implicit in the diagnosis.']),
  NtQbin._make(['pathogenicity', 'PY', 'PATHOGEN', 'Used with microorganisms, viruses, and parasites for studies of their ability to cause disease in man, animals, or plants.']),
  NtQbin._make(['pathology', 'PA', 'PATHOL', 'Used for organ, tissue, or cell structure in disease states.']),
  NtQbin._make(['pharmacokinetics', 'PK', 'PHARMACOKIN', 'Used for the mechanism, dynamics and kinetics of exogenous chemical and drug absorption, biotransformation, distribution, release, transport, uptake and elimination as a function of dosage, extent and rate of metabolic processes.']),
  NtQbin._make(['pharmacology', 'PD', 'PHARMACOL', 'Used with drugs and exogenously administered chemical substances for their effects on living tissues and organisms. It includes acceleration and inhibition of physiological and biochemical processes and other pharmacologic mechanisms of action.']),
  NtQbin._make(['physiology', 'PH', 'PHYSIOL', 'Used with organs, tissues, and cells of unicellular and multicellular organisms for normal function. It is used also with biochemical substances, endogenously produced, for their physiologic role.']),
  NtQbin._make(['physiopathology', 'PP', 'PHYSIOPATHOL', 'Used with organs and diseases for disordered function in disease states.']),
  NtQbin._make(['poisoning', 'PO', 'POIS', 'Used with drugs, chemicals, and industrial materials for human or animal poisoning, acute or chronic, whether the poisoning is accidental, occupational, suicidal, by medication error, or by environmental exposure.']),
  NtQbin._make(['prevention & control', 'PC', 'PREV', 'Used with disease headings for increasing human or animal resistance against disease (e.g., immunization), for control of transmission agents, for prevention and control of environmental hazards, or for prevention and control of social factors leading to disease. It includes preventive measures in individual cases.']),
  NtQbin._make(['psychology', 'PX', 'PSYCHOL', 'Used with non-psychiatric diseases, techniques, and named groups for psychologic, psychiatric, psychosomatic, psychosocial, behavioral, and emotional aspects, and with psychiatric disease for psychologic aspects; used also with animal terms for animal behavior and psychology.']),
  NtQbin._make(['radiation effects', 'RE', 'RAD EFF', 'Used for effects of ionizing and nonionizing radiation upon living organisms, organs and tissues, and their constituents, and upon physiologic processes. It includes the effect of irradiation on drugs and chemicals.']),
  NtQbin._make(['radiotherapy', 'RT', 'RADIOTHER', 'Used with disease headings for the therapeutic use of ionizing and nonionizing radiation. It includes the use of radioisotope therapy.']),
  NtQbin._make(['rehabilitation', 'RH', 'REHABIL', 'Used with diseases and surgical procedures for restoration of function of the individual.']),
  NtQbin._make(['secondary', 'SC', 'SECOND', 'Used with neoplasms to indicate the secondary location to which the neoplastic process has metastasized.']),
  NtQbin._make(['standards', 'ST', 'STAND', 'Used with facilities, personnel, and program headings for the development, testing, and application of standards of adequacy or acceptable performance and with chemicals and drugs for standards of identification, quality, and potency. It includes health or safety standards in industries and occupations.']),
  NtQbin._make(['statistics & numerical data', 'SN', 'STATIST', 'Used with non-disease headings for the expression of numerical values that describe particular sets or groups of data. It includes level of use of equipment and supplies, facilities and services and procedures and techniques. It excludes supply or demand for which "supply & distribution" is used']),
  NtQbin._make(['supply & distribution', 'SD', 'SUPPLY', 'Used for the quantitative availability and distribution of material, equipment, health services, personnel, and facilities. It excludes food supply and water supply in industries and occupations.']),
  NtQbin._make(['surgery', 'SU', 'SURG', 'Used for operative procedures on organs, regions, or tissues in the treatment of diseases, including tissue section by lasers. It excludes transplantation, for which "transplantation" is used.']),
  NtQbin._make(['therapeutic use', 'TU', 'THER USE', 'Used with drugs, biological preparations, and physical agents for their use in the prophylaxis and treatment of disease. It includes veterinary use.']),
  NtQbin._make(['therapy', 'TH', 'THER', 'Used with diseases for therapeutic interventions except drug therapy, diet therapy, radiotherapy, and surgery, for which specific subheadings exist. The concept is also used for articles and books dealing with multiple therapies.']),
  NtQbin._make(['toxicity', 'TO', 'TOX', 'Used with drugs and chemicals for experimental human and animal studies of their ill effects. It includes studies to determine the margin of safety or the reactions accompanying administration at various dose levels. It is used also for exposure to environmental agents. Poisoning should be considered for life-threatening exposure to environmental agents.']),
  NtQbin._make(['transmission', 'TM', 'TRANSM', 'Used with diseases for studies of the modes of transmission.']),
  NtQbin._make(['transplantation', 'TR', 'TRANSPL', 'Used with organs, tissues, or cells for transplantation from one site to another within the same subject, or from one subject to another of the same species or different species.']),
  NtQbin._make(['trends', 'TD', 'TRENDS', 'Used for the manner in which a subject changes, qualitatively or quantitatively, with time, whether past, present, or future. It excludes discussions of the course of disease in particular patients.']),
  NtQbin._make(['ultrastructure', 'UL', 'ULTRASTRUCT', 'Used with tissues and cells (including neoplasms) and microorganisms for microanatomic structures, generally below the size visible by light microscopy.']),
  NtQbin._make(['urine', 'UR', 'URINE', 'Used for the presence or analysis of substances in the urine, and also for the examination of, or changes in, the urine in disease.']),
  NtQbin._make(['veterinary', 'VE', 'VET', 'Used for naturally occurring diseases in animals, or for diagnostic, preventive, or therapeutic procedures used in veterinary medicine.']),
  NtQbin._make(['virology', 'VI', 'VIROL', 'Used with organs, animals, and higher plants and with diseases for virologic studies. For bacteria, rickettsia, and fungi, "microbiology" is used; for parasites, "parasitology" is used.']),
]

# Copyright (C) 2014-2020 DV Klopfenstein. All rights reserved



"""Fetch items and write"""

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2016-present DV Klopfenstein. All rights reserved."
__license__ = "GPL"

from pmidcite.cfg import Cfg
from pmidcite.eutils.cmds.base import EntrezUtilities


#### class EntrezCommands(EntrezUtilities):
class CommandBase(EntrezUtilities):
    """Fetch and write text"""

    def __init__(self, retmax=10000, rettype='medline', retmode='text', batch_size=100, **kws):
        kws_base = {k:v for k, v in kws.items() if k in EntrezUtilities.exp_kws}
        cfg = Cfg()
        super(CommandBase, self).__init__(
            cfg.get_email(), cfg.get_apikey(), cfg.get_tool(), **kws_base)
        self.batch_size = batch_size
        self.retmax = retmax
        self.rettype = rettype
        self.retmode = retmode


# Copyright (C) 2016-present DV Klopfenstein. All rights reserved.



"""Groups papers using the NIH percentile"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from collections import namedtuple


class NihGrouper:
    """Groups papers using the NIH percentile"""

    ntobj = namedtuple('NtNihGroup', 'val txt')

    def __init__(self, group1_min=2.1, group2_min=15.7, group3_min=83.9, group4_min=97.5):
        self.min1 = group1_min
        self.min2 = group2_min
        self.min3 = group3_min
        self.min4 = group4_min

    def get_group(self, nih_percentile):
        """Assign group numbers to the NIH percentile values using the 68-95-99.7 rule"""
        # No NIH percentile yet assigned. This paper should be checked out.
        if nih_percentile is None or nih_percentile == -1:
            return 5
        #  2.1% -3 SD: Very low citation rate
        if nih_percentile < self.min1:  # default: 2.1
            return 0
        # 13.6% -2 SD: Low citation rate
        if nih_percentile < self.min2:  # default: 15.7
            return 1
        # 68.2% -1 SD to +1 SD: Average citation rate
        if nih_percentile < self.min3:  # default: 83.9
            return 2
        # 13.6% +2 SD: High citation rate
        if nih_percentile < self.min4:  # default: 97.5
            return 3
        #  2.1% +3 SD: Very high citation rate
        return 4

    def add_arguments(self, parser):
        """Add NIH grouper arguments to the parser"""
        # pylint: disable=line-too-long
        parser.add_argument(
            '-1', metavar='group1_min', dest='min1', default=self.min1, type=float,
            help='Minimum NIH percentile to be placed in group 1 (default: {D})'.format(D=self.min1))
        parser.add_argument(
            '-2', metavar='group2_min', dest='min2', default=self.min2, type=float,
            help='Minimum NIH percentile to be placed in group 2 (default: {D})'.format(D=self.min2))
        parser.add_argument(
            '-3', metavar='group3_min', dest='min3', default=self.min3, type=float,
            help='Minimum NIH percentile to be placed in group 3 (default: {D})'.format(D=self.min3))
        parser.add_argument(
            '-4', metavar='group4_min', dest='min4', default=self.min4, type=float,
            help='Minimum NIH percentile to be placed in group 4 (default: {D})'.format(D=self.min4))

    def get_list(self):
        """Get the dividing values as a list"""
        return [self.min1, self.min2, self.min3, self.min4]

    def __str__(self):
        """Print NIH percentiles which divide pmidcite groups"""
        return 'Group min percentiles: 1({}) 2({}) 3({}) 4({})'.format(
            self.min1, self.min2, self.min3, self.min4)


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



"""Plot the types of content and their amount in PubMed"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import matplotlib as mpl
mpl.use('agg')
import matplotlib.pyplot as plt

from pmidcite.eutils.pubmed.counts.data import DataMgr


class PubMedPlot:
    """Plot the types of content and their amount in PubMed"""

    arrow_p = {
        'color': 'k',
        'linewidth':0.4,
        'shape':'full',
        'length_includes_head':True,
        'head_width':.3,
        'head_length':400000,
        'head_starts_at_zero': False,
        'overhang':0.0,
        # https://matplotlib.org/3.1.1/gallery/lines_bars_and_markers/joinstyle.html
        'capstyle':'butt',    # butt round projecting
        'joinstyle':'bevel',  # miter, round, bevel
    }

    def __init__(self, name2cnt):
        self.name2cnt = name2cnt
        self.dataobj = DataMgr(self.name2cnt)

    def plt_content_counts(self, fout_png, dpi=800):
        """Plot pubmed content"""
        a2n = self.name2cnt
        xmax = a2n['all']
        ymax = 17.5
        # Remove automatically-added 5% axes padding
        mpl.rcParams['axes.autolimit_mode'] = 'data'  # 'data' or 'round_numbers'
        mpl.rcParams['axes.xmargin'] = 0
        mpl.rcParams['axes.ymargin'] = 0
        # Get figure and axes with axes turned off and axes whitespace removed
        fig, axes = plt.subplots()
        fig.set_size_inches(6.4, 1.8)
        axes.get_xaxis().set_visible(False)
        axes.get_yaxis().set_visible(False)
        axes.set_frame_on(False)
        axes.set_xlim(0, xmax+500000)
        axes.set_ylim(0, ymax+1)
        axes.grid(False)
        # Add horizontal bars for: PubMed, PMC, and other
        bbars = []
        for xvals, yval, dct in self.dataobj.get_pubmed_colorbars(ymax-5):
            print('XY-COORD', xvals, yval)
            bbars.append(axes.broken_barh(xvals, yval, **dct))
        # Add text to colorbars using xy coordinates from 'XY-COORD'
        self._text_on_bars(axes)
        # Add Dimension lines
        # https://schoolworkhelper.net/technical-drawing-alphabet-of-line/
        self._add_bounding_lines_all(xmax, ymax)
        self._add_bounding_lines_medline(a2n['medline_n_inprocess'], ymax-2)
        self._add_bounding_lines_pmc(ymax-6)
        self._add_bounding_lines_other(a2n['all_ml0_pmc0'], ymax-10, xmax)
        self._add_bounding_lines_pmc_100(ymax-12)
        # Add legend
        axes.legend(loc='lower left', fontsize=8, ncol=2,
                    bbox_to_anchor=(0.015, 0.0), borderaxespad=0.1,
                    handletextpad=.2, columnspacing=1.0, labelspacing=.2)
        # Save figure
        plt.savefig(fout_png, bbox_inches='tight', pad_inches=0, dpi=dpi)
        print('  WROTE: {PNG}'.format(PNG=fout_png))

    @staticmethod
    def _text_on_bars(axes):
        """Add text on colorbars using xy coordinates from 'XY-COORD'"""
        dct_txt = {'va':'bottom', 'fontweight':'bold'}
        # bar text: MEDLINE
        # XY-COORD [(       0, 22956376)] (11.5, 1.8)
        # XY-COORD [(22956376,   558119)] (11.5, 1.8)
        # XY-COORD [(23514495,  3505412)] (11.5, 1.8)
        axes.text(22956376*.62, 11.5, 'MEDLINE', fontsize=9, color='white', ha='center', **dct_txt)
        axes.text(22956376*.86, 11.5, 'database', fontsize=9, color='white', ha='center', **dct_txt)
        axes.text(22956376 + 558119 + 3505412*.32, 11.5, '(db)', fontsize=9, color='white', ha='center', **dct_txt)
        # bar text: PMC
        # XY-COORD [(23378342,  136153)] (9.5, 1.8)
        # XY-COORD [(23514495, 3505412)] (9.5, 1.8)
        # XY-COORD [(27019907, 1687687)] (9.5, 1.8)
        axes.text(23514495 + 3505412*.66, 9.5, 'PMC', fontsize=9, color='white', ha='center', **dct_txt)
        axes.text(27019907 + 1687687*.43, 9.5, 'db', fontsize=9, color='white', ha='center', **dct_txt)
        # bar text: other
        # XY-COORD [(28707594, 1819828)] (5.5, 1.8)
        axes.text(28707594+ 1819828/2.0, 5.7, 'Other', fontsize=6.3, color='black', ha='center', **dct_txt)

    def chk_content_counts(self):
        """Check the content typename and the count of that type"""
        a2n = self.name2cnt
        assert a2n['nihms_pub0'] + a2n['nihms_pub1'] == a2n['nihms']
        assert a2n['pmcsd_pmc0'] + a2n['pmcsd_pmc1'] == a2n['pmcsd']
        assert a2n['medline_pmc0'] + a2n['medline_pmc1'] == a2n['medline_all']
        #    557,390 inprocess_A_all      inprocess[sb]
        #    422,103 inprocess_A_pmc0     inprocess[sb] NOT pubmed pmc[sb]
        #    135,287 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
        #    557,390 inprocess_ml0        inprocess[sb] NOT medline[sb]
        #          0 inprocess_ml1        inprocess[sb] AND medline[sb]
        assert a2n['inprocess_A_all'] == a2n['inprocess_A_pmc0'] + a2n['inprocess_A_pmc1']
        # 30,514,237 all          all [sb]
        #  1,818,474 all_ml0_pmc0 all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb]
        # 28,695,763 ml1_pmc1     inprocess[sb] OR medline[sb] OR pubmed pmc[sb]
        assert a2n['all'] == a2n['all_ml0_pmc0'] + a2n['ml1_pmc1']
        #  3,116,339 pmnml_A              pubmednotmedline[sb]
        #  3,116,339 pmnml_B              pubmednotmedline[sb] NOT medline[sb]
        #  3,116,339 pmnml_C_ip0          pubmednotmedline[sb] NOT inprocess[sb]
        #          0 pmnml_0_ip1          pubmednotmedline[sb] AND inprocess[sb]
        #  1,612,274 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
        #  1,504,065 pmnml_A_pmc0         pubmednotmedline[sb] NOT pubmed pmc[sb]
        assert a2n['pmnml_A'] == a2n['pmnml_A_pmc1'] + a2n['pmnml_A_pmc0']
        #  5,323,939 pmc_all              pubmed pmc[sb]
        #    135,287 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
        #  3,503,057 medline_pmc1         medline[sb] AND pubmed pmc[sb]
        #  1,612,274 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
        pmc_notmedline = a2n['pmc_all'] - a2n['medline_pmc1'] - a2n['inprocess_A_pmc1']
        print('  {N:10,} of {M:10,} PMC (not MEDLINE)'.format(M=a2n['pmc_all'], N=pmc_notmedline))
        print(pmc_notmedline - a2n['pmnml_A_pmc1'])
        print('  {N:10,} of {M:10,} PubMed(not MEDLINE, PMC)'.format(
            N=a2n['all_ml0_pmc0'], M=a2n['all']))
        print('  {T:10,} = {A:10,} (MEDLINE OR PMC) + {B:10,} (not MEDLINE OR PMC)'.format(
            T=a2n['ml1_pmc1'] + a2n['all_ml0_pmc0'],
            A=a2n['ml1_pmc1'],
            B=a2n['all_ml0_pmc0']))
        total = a2n['all']
        au_all = a2n['au_all']
        print('  {N:10,} of {M:10,} {P:3.5f}% author ms'.format(
            N=au_all, M=total, P=100.0*au_all/total))
        print('  {T:10,} = {A:10,} (MEDLINE OR PMC) + {B:10,} (not MEDLINE OR PMC)'.format(
            T=a2n['ml1_pmc1'] + a2n['all_ml0_pmc0'],
            A=a2n['ml1_pmc1'],
            B=a2n['all_ml0_pmc0']))

    def _add_bounding_lines_all(self, xend, ymax):
        """Add bounding lines"""
        plt.plot((0, 0), (ymax-12, ymax+1), color='k', linewidth=0.4)        # LEFT  PubMed LINE
        plt.plot((xend, xend), (ymax-12, ymax+1), color='k', linewidth=0.4)  # RIGHT PubMed LINE
        plt.arrow(4200000, ymax, -4200000, 0, **self.arrow_p)
        plt.arrow(27600000, ymax, xend-27600000, 0, **self.arrow_p)
        ntd = self.dataobj.pltdata_pubmed['PubMed']
        txt = '~{N:4.1f} million (M) citations indexed by PubMed'.format(
            N=round(ntd.count/1000000.0, 1))
        plt.annotate(txt, (4400000, ymax-.5), fontweight='bold')

    def _add_bounding_lines_medline(self, xend, yval):
        """Add bounding lines"""
        plt.plot((xend, xend), (yval-8.7, yval), color='k', linewidth=0.4)   # BLUE-YELLOW DIVIDER
        plt.plot((xend, xend), (yval-14, yval-9.3), color='k', linewidth=0.4)  # BLUE-YELLOW DIVIDER
        plt.arrow(4200000, yval-1, -4200000, 0, **self.arrow_p)
        plt.arrow(16000000, yval-1, xend-16000000, 0, **self.arrow_p)
        ntd = self.dataobj.pltdata_pubmed['MEDLINE_n_inprocess']
        txt = '~{N:4.1f}M ({P:4.1f}%) MEDLINE'.format(N=round(ntd.count/1000000.0, 1), P=ntd.perc)
        plt.annotate(txt, (4400000, yval-1.5))

    def _add_bounding_lines_pmc(self, yval):
        """Add bounding lines"""
        a2n = self.name2cnt
        # pylint: disable=line-too-long
        #   # PMC
        #   ([(ml2, a2n['inprocess_A_pmc1'])], (23, 1.8), {'facecolors':'tab:cyan', **par}),
        #   ([(ml3, a2n['medline_pmc1'])],     (23, 1.8), {'facecolors':'tab:blue', **par}),
        #   ([(ml4, a2n['pmnml_A_pmc1'] + a2n['pmc_unknown'])], (23, 1.8), {'facecolors':'y', **par}),
        pmc_x0 = a2n['medline_pmc0'] + a2n['inprocess_A_pmc0']
        pmc_all = a2n['pmc_all']
        pmc_xn = pmc_x0 + pmc_all
        plt.plot((pmc_x0, pmc_x0), (yval-2.7, yval+2), color='k', linewidth=0.4)    # UPPER CYAN DIVIDER
        plt.plot((pmc_x0, pmc_x0), (yval-4.7, yval-3.3), color='k', linewidth=0.4)  # UPPER CYAN DIVIDER
        # PMC
        plt.arrow(pmc_x0-9500000, yval-1, 9500000, 0, **self.arrow_p)
        plt.arrow(pmc_xn+1300000, yval-1, -1300000, 0, **self.arrow_p)
        ntd = self.dataobj.pltdata_pubmed['PMC']
        txt = '~{N:5.1f}M ({P:3.1f}%) PMC'.format(N=round(ntd.count/1000000.0, 1), P=ntd.perc)
        plt.annotate(txt, (4400000, yval-1.5))
        # PMC Only
        plt.arrow(a2n['medline_n_inprocess']-11000000, yval-3, 11000000, 0, **self.arrow_p)
        plt.arrow(pmc_xn+1300000, yval-3, -1300000, 0, **self.arrow_p)
        ntd = self.dataobj.pltdata_pubmed['PMC_only']
        txt = '~{N:5.1f}M (  {P:3.1f}%) PMC only'.format(N=round(ntd.count/1000000.0, 1), P=ntd.perc)
        plt.annotate(txt, (4400000, yval-3.5))

    def _add_bounding_lines_other(self, other_sz, yval, xmax):
        """Add bounding lines"""
        xval = xmax - other_sz
        plt.plot((xval, xval), (yval-8, yval+4), color='k', linewidth=0.4)  # YELLOW-ORANGE DIVIDER
        plt.arrow(xval-14200000, yval-1, 14200000, 0, **self.arrow_p)
        plt.arrow(xmax+600000, yval-1, -600000, 0, **self.arrow_p)
        ntd = self.dataobj.pltdata_pubmed['other']
        txt = '~{N:5.1f}M ({P:5.1f}%) Other'.format(N=round(ntd.count/1000000.0, 1), P=ntd.perc)
        plt.annotate(txt, (4400000, yval-1.5))

    def _add_bounding_lines_pmc_100(self, yval):
        """Add bounding lines"""
        a2n = self.name2cnt
        # pylint: disable=line-too-long
        #   # PMC
        #   ([(ml2, a2n['inprocess_A_pmc1'])], (23, 1.8), {'facecolors':'tab:cyan', **par}),
        #   ([(ml3, a2n['medline_pmc1'])],     (23, 1.8), {'facecolors':'tab:blue', **par}),
        #   ([(ml4, a2n['pmnml_A_pmc1'] + a2n['pmc_unknown'])], (23, 1.8), {'facecolors':'y', **par}),
        pmc_all = a2n['pmc_all']
        pmc_ml1 = a2n['medline_pmc1'] + a2n['inprocess_A_pmc1']
        pmc_ml0 = pmc_all - pmc_ml1
        pmc_x0 = a2n['medline_pmc0'] + a2n['inprocess_A_pmc0']
        pmc_x1 = pmc_x0 + pmc_ml1
        pmc_xn = pmc_x0 + pmc_all
        plt.plot((pmc_x0, pmc_x0), (yval-6, yval+0.7), color='k', linewidth=0.4)  # LOWER CYAN DIVIDER
        # PMC AND MEDLINE
        plt.annotate('MEDLINE', (pmc_x0+pmc_ml1/2.0, yval-2.0), ha='center', va='center', fontsize=8)
        # PMC|MEDLINE
        plt.arrow(pmc_x0+900000, yval-3.3, -900000, 0, **self.arrow_p)
        plt.arrow(pmc_x1-900000, yval-3.3, 900000, 0, **self.arrow_p)
        plt.arrow(pmc_xn+700000, yval-3.3, -700000, 0, **self.arrow_p)
        txt_ml1 = '{P:2.0f}%'.format(P=round(100.0*pmc_ml1/pmc_all))
        plt.annotate(txt_ml1, (pmc_x0+pmc_ml1/2.0, yval-3.3), ha='center', va='center', fontsize=8)
        txt_ml0 = '{P:2.0f}%'.format(P=round(100.0*(pmc_all-pmc_ml1)/pmc_all))
        plt.annotate(txt_ml0, (pmc_x1 + pmc_ml0/2.0, yval-3.3), ha='center', va='center', fontsize=8)
        # PMC
        plt.arrow(pmc_x0+1600000, yval-5, -1600000, 0, **self.arrow_p)
        plt.arrow(pmc_xn-1600000, yval-5, 1600000, 0, **self.arrow_p)
        plt.annotate('PMC', (pmc_x0+pmc_all/2.0, yval-5), ha='center', va='center')


# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Plot the types of content and their amount in PubMed"""

# To check whether a journal is indexed in MEDLINE and stored in PMC,
# search the NLM Catalog:
# 
#     https://www.ncbi.nlm.nih.gov/nlmcatalog
# 
# PMC also maintains a journals list, which is dynamic list, 
# that provides journal participation level information:
# 
#     https://www.ncbi.nlm.nih.gov/pmc/journals/
#
# Lidia Hutcherson
# LinkOut Development Team
# Information Engineering Branch
# National Center for Biotechnology Information
# US National Library of Medicine
# Linking to a World of Resources
# http://www.ncbi.nlm.nih.gov/projects/linkout/

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import sys
import datetime
import collections as cx
from pmidcite.eutils.cmds.base import EntrezUtilities


# pylint: disable=line-too-long
class PubMedDnld(EntrezUtilities):
    """Describe the percentage of various types of content in PubMed"""

    # https://pubmed.ncbi.nlm.nih.gov/help/#citation-status-subsets
    name2query = cx.OrderedDict([
        ('all', 'all [sb]'),
        ('all_ml0_pmc0', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb]'),
        ('all_ml0_pmc0_free0', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT free full text[sb]'),
        ('all_ml0_pmc0_free1', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] AND free full text[sb]'),
        ('all_ml0_pmc0b', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd'),
        # Categories include: "Online, ahead of print" and old articles/books, Reviews found in GeneReviews, etc.
        ('all_ml0_pmc0d', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd AND pubmednotmedline[sb]'),
        ('all_ml0_pmc0c', 'all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pubmednotmedline[sb]'),
        ('ml1_pmc1', 'inprocess[sb] OR medline[sb] OR pubmed pmc[sb]'),
        # Over 2 million FREE-text articles in PubMed that are NOT in PMC
        ('pm1free_pmc0', 'free full text[sb] NOT loprovpmc[sb]'),
        # Over 2 million FREE-text articles in PubMed that are NOT in PMC
        ('ml1free_pmc0', 'free full text[sb] AND medline[sb] NOT pubmed pmc[sb]'),
        # publisher[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT # pmcbook:
        #   1. Citations recently added to PubMed via electronic submission
        #   from a publisher, and are soon to proceed to the next stage,
        #   PubMed - in process (see below).
        #   2. Also for citations received
        #   before late 2003 if they are from journals not indexed for
        #   MEDLINE, or from a journal that was accepted for MEDLINE
        #   after the citations' publication date.
        #   NOTE: These citations bibliographic data have not been reviewed.
        ('pub_init0', 'publisher[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook'),
        ('pub_init1', 'publisher[sb] AND (pubstatusnihms OR pubstatuspmcsd OR pmcbook)'),
        ('pub', 'publisher[sb]'),
        # pubstatusnihms AND publisher[sb]:
        #   Author manuscripts submitted to PMC that fall under the NIH Public Access Policy.
        # https://www.ncbi.nlm.nih.gov/pmc/about/submission-methods/
        #   Some publishers will initiate the NIHMS deposit process for an author,
        #   but the author must complete two approval steps to load the PMC-ready version to PMC
        ('nihms_pub0', 'pubstatusnihms AND publisher[sb]'),
        ('nihms_pub1', 'pubstatusnihms NOT publisher[sb]'),
        ('nihms', 'pubstatusnihms'),
        ('nihms_pmc1', 'pubstatusnihms AND pubmed pmc[sb]'),
        ('nihms_pmc0', 'pubstatusnihms NOT pubmed pmc[sb]'),
        # pubstatuspmcsd AND publisher[sb]:
        #   Records for selective deposit articles in PMC.
        #   These are articles published in non-MEDLINE journals
        #   where the publisher has chosen to deposit in PMC
        #   only those articles that fall under the NIH Public Access Policy.
        ('pub_pmcsd', 'pubstatuspmcsd AND publisher[sb]'),
        ('pmcsd', 'pubstatuspmcsd'),
        ('pmcsd_and_nihms', 'pubstatuspmcsd AND pubstatusnihms'),  # DISJOINT
        ('pmcsd_pmc1', 'pubstatuspmcsd AND pubmed pmc[sb]'),
        ('pmcsd_pmc0', 'pubstatuspmcsd NOT pubmed pmc[sb]'),
        # inprocess MEDLINE:  inprocess[sb]
        #   MeSH terms will be assigned if the subject of the article is within the scope of MEDLINE.
        ('inprocess_A_all', 'inprocess[sb]'),
        ('inprocess_A_pmc0', 'inprocess[sb] NOT pubmed pmc[sb]'),
        ('inprocess_A_pmc1', 'inprocess[sb] AND pubmed pmc[sb]'),
        ('inprocess_ml0', 'inprocess[sb] NOT medline[sb]'),
        ('inprocess_ml1', 'inprocess[sb] AND medline[sb]'),
        ('medline_n_inprocess', 'medline[sb] OR inprocess[sb]'),
        # medline[sb]:
        #   Citations that have been indexed with MeSH terms, Publication Types, Substance Names, etc.
        ('medline_all', 'medline[sb]'),
        ('medline_pmc0', 'medline[sb] NOT pubmed pmc[sb]'),
        ('medline_pmc1', 'medline[sb] AND pubmed pmc[sb]'),
        # pubmed pmc[sb]:
        #   Citations in PMC
        ('pmc_all', 'pubmed pmc[sb]'),
        #  5,323,939 pmc_all              pubmed pmc[sb]
        #    135,287 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
        #  3,503,057 medline_pmc1         medline[sb] AND pubmed pmc[sb]
        #  1,612,274 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
        ('pmc_all?', 'pubmed pmc[sb] AND (inprocess[sb] OR medline[sb] OR pubmednotmedline[sb])'),
        ('pmc_m1', 'pubmed pmc[sb] AND (inprocess[sb] OR medline[sb])'),
        ('pmc_m0', 'pubmed pmc[sb] NOT inprocess[sb] NOT medline[sb]'),
        ('pmc_m0b', 'pubmed pmc[sb] AND pubmednotmedline[sb]'),
        ('pmc_unknown', 'pubmed pmc[sb] NOT inprocess[sb] NOT medline[sb] NOT pubmednotmedline[sb]'),
        # https://www.ncbi.nlm.nih.gov/pmc/about/submission-methods/
        ('au_all', 'author manuscript all[sb]'),
        ('au_all1', 'author manuscript all[sb] AND all[sb]'),
        ('au_ss', 'author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook'),
        ('au_ss_m0', 'author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb]'),
        ('au_ss_ml0', 'author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb] NOT inprocess[sb]'),
        ('au_ss_pmconly', 'author manuscript[sb] AND pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb] NOT inprocess[sb]'),
        ('au_nihms', 'author manuscript[sb] AND pubstatusnihms'),
        ('au_pmcsd', 'author manuscript[sb] AND pubstatuspmcsd'),
        ('au_pmcbook', 'author manuscript[sb] AND pmcbook'),
        ('au_ml1', 'author manuscript[sb] AND (medline[sb] OR inprocess[sb])'),
        ('au_ml0', 'author manuscript[sb] NOT medline[sb] NOT inprocess[sb]'),
        ('au_ml0_pmc0', 'author manuscript[sb] NOT medline[sb] NOT inprocess[sb] NOT pubmed pmc[sb]'),
        ('au_ml0_pmc1', 'author manuscript[sb] AND pubmed pmc[sb] NOT medline[sb] NOT inprocess[sb]'),
        ('au_m1', 'author manuscript[sb] AND medline[sb]'),
        ('au_ip1', 'author manuscript[sb] AND inprocess[sb]'),
        ('au_ml1_or_pmc1', 'author manuscript[sb] AND (medline[sb] OR pubmed pmc[sb])'),
        ('au_ml1_and_pmc1', 'author manuscript[sb] AND medline[sb] AND pubmed pmc[sb]'),
        ('au_ml1_and_pmc1_or_ip1', 'author manuscript[sb] AND (inprocess[sb] OR medline[sb] AND pubmed pmc[sb])'),
        ('au_pmc1', 'author manuscript[sb] AND pubmed pmc[sb]'),
        ('au_pmc1_ip1', 'author manuscript[sb] AND pubmed pmc[sb] AND inprocess[sb]'),
        ('au_pmc1_ml0', 'author manuscript[sb] AND pubmed pmc[sb] NOT medline[sb]'),
        ('au_pmc0', 'author manuscript[sb] NOT pubmed pmc[sb]'),
        # 7,680 embargoed
        ('au_embargo', 'author manuscript[sb] AND medline[sb] NOT pubmed pmc[sb]'),
        # pubmednotmedline[sb]
        #   Citations that will not receive MEDLINE indexing because:
        #    * they are for articles in non-MEDLINE journals, or
        #    * they are for articles in MEDLINE journals but the articles are out of scope, or
        #    * they are from issues published prior to the date the journal was selected for indexing, or
        #    * citations to articles from journals that deposit their full text articles in PMC
        #      but have not yet been recommended for indexing in MEDLINE.
        ('pmnml_A', 'pubmednotmedline[sb]'),
        ('pmnml_B', 'pubmednotmedline[sb] NOT medline[sb]'),
        ('pmnml_C', 'pubmednotmedline[sb] NOT pubmed pmc[sb]'),
        ('pmnml_D', 'pubmednotmedline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms[sb] NOT pubstatuspmcsd[sb]'),
        ('pmnml_C_ip0', 'pubmednotmedline[sb] NOT inprocess[sb]'),
        ('pmnml_0_ip1', 'pubmednotmedline[sb] AND inprocess[sb]'),
        ('pmnml_A_pmc1', 'pubmednotmedline[sb] AND pubmed pmc[sb]'),
        ('pmnml_A_pmc0', 'pubmednotmedline[sb] NOT pubmed pmc[sb]'),
        # pmcbook:
        #   Book and book chapter citations available on the NCBI Bookshelf.
        ('pmcbook_A', 'pmcbook'),
        ('pmcbook_A_ml0', 'pmcbook NOT medline[sb]'),
        ('pmcbook_A_pmc0', 'pmcbook NOT pubmed pmc[sb]'),
        ('pmcbook_A_pmc0', 'pmcbook NOT pubmed pmc[sb]'),
    ])

    def __init__(self, email, apikey, tool):
        super(PubMedDnld, self).__init__(email, apikey, tool)
        self.date = str(datetime.datetime.now().date()).replace('-', '_')

    def get_content_counts(self, file_py, force_dnld):
        """Read or download the PubMed content counts"""
        if not os.path.exists(file_py) or force_dnld:
            name2cnt = self.dnld_content_counts()
            self.wrpy_count_data(file_py, name2cnt)
            self.chk_content_counts(name2cnt)
            return name2cnt, self.date
        return self.load_count_data(file_py)

    @staticmethod
    def load_count_data(fin_py):
        """Load NIH iCite information from Python modules"""
        import importlib
        spec = importlib.util.spec_from_file_location("module.name", fin_py)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        print('  READ: {PY}'.format(PY=fin_py))
        return mod.CNTS, mod.DATE

    def dnld_content_counts(self):
        """Get the count of various types of content in PubMed"""
        name2cnt = {}
        for name, query in self.name2query.items():
            cnt = self.dnld_count(query)
            print('  Downloaded {N:12,} {A:20} {Q}'.format(N=cnt, A=name, Q=query))
            name2cnt[name] = cnt
        # 2019/01/04 PubMed Query returns 30,500,360 results: all [sb]
        # 2019/01/04 PubMed Query returns  3,499,063 results: medline[sb] AND pubmed pmc[sb]
        # 2019/01/04 PubMed Query returns  3,636,767 results: medline[sb] OR inprocess[sb] AND pubmed pmc[sb]
        # 2019/01/04 PubMed Query returns  5,321,522 results: pubmed pmc[sb]
        # 2019/01/04 PubMed Query returns  3,499,063 results: pubmed pmc[sb] AND medline[sb]
        # 2019/01/04 PubMed Query returns    137,704 results: pubmed pmc[sb] AND inprocess[sb]
        # 2019/01/04 PubMed Query returns  3,636,767 results: pubmed pmc[sb] AND (inprocess[sb] OR medline[sb])
        # 2019/01/04 PubMed Query returns  1,229,362 results: pubmed pmc[sb] AND pubmednotmedline[sb]
        # 2019/01/04 PubMed Query returns  5,321,522 results: pubmed pmc[sb]

        # To search for the total number of citations in PubMed that are PubMed only citations
        # (pubmednotmedline[sb]-citations that will not receive MEDLINE indexing because
        #     * they are for articles in non-MEDLINE journals, or
        #     * they are for articles in MEDLINE journals but the articles are out of scope, or
        #     * they are from issues published prior to the date the journal was selected for indexing, or
        #     * citations to articles from journals that deposit their full text articles in PMC but have not yet been recommended for indexing in MEDLINE),
        #     * search for pubmednotmedline[sb] in the search box:
        #       pubmednotmedline[sb]
        # 2019/01/04 PubMed Query returns    422,513 results: inprocess[sb] NOT pubmed pmc[sb]
        # 2019/01/04 PubMed Query returns    137,704 results: inprocess[sb] AND pubmed pmc[sb]
        #%%% % PubMed Query: author manuscript[sb] AND pubmed pmc[sb]
        #%%% % PubMed Query: pubmed pmc[sb]
        # PubMed Query: pubmed pmc[sb] AND pubmednotmedline[sb]
        # PubMed Query: pubmed pmc[sb] AND pubmednotmedline[sb]
        # PubMed Query: author manuscript[sb] AND pubmed pmc[sb]
        return name2cnt

    def wrpy_count_data(self, fout_py, name2cnt):
        """Write PubMed count data into a Python module"""
        with open(fout_py, 'w') as prt:
            prt.write('"""Downloaded PubMed count data"""\n\n')
            prt.write("DATE = '{DATE}'\n\n".format(DATE=self.date))
            prt.write('# pylint: disable=line-too-long\n')
            self.prt_content_counts(name2cnt, prt)
            prt.write('\n')
            self.prt_content_cntdct(name2cnt, prt)
            print('  WROTE: {PY}'.format(PY=fout_py))

    def prt_content_counts(self, name2cnt, prt=sys.stdout):
        """Print the content typename and the count of that type"""
        for name, query in self.name2query.items():
            if name in name2cnt:
                cnt = name2cnt[name]
                prt.write('# {N:10,} {NAME:20} {Q}\n'.format(N=cnt, NAME=name, Q=query))

    def prt_content_cntdct(self, name2cnt, prt=sys.stdout):
        """Print the content typename and the count of that type"""
        prt.write('CNTS = {\n')
        for name in self.name2query:
            cnt = name2cnt[name]
            prt.write('    "{NAME}": {N},\n'.format(N=cnt, NAME=name))
        prt.write('}\n')

    @staticmethod
    def chk_content_counts(a2n):
        """Check the content typename and the count of that type"""
        assert a2n['nihms_pub0'] + a2n['nihms_pub1'] == a2n['nihms']
        assert a2n['pmcsd_pmc0'] + a2n['pmcsd_pmc1'] == a2n['pmcsd']
        assert a2n['medline_pmc0'] + a2n['medline_pmc1'] == a2n['medline_all']
        #    557,390 inprocess_A_all      inprocess[sb]
        #    422,103 inprocess_A_pmc0     inprocess[sb] NOT pubmed pmc[sb]
        #    135,287 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
        #    557,390 inprocess_ml0        inprocess[sb] NOT medline[sb]
        #          0 inprocess_ml1        inprocess[sb] AND medline[sb]
        assert a2n['inprocess_A_all'] == a2n['inprocess_A_pmc0'] + a2n['inprocess_A_pmc1']
        # 30,514,237 all          all [sb]
        #  1,818,474 all_ml0_pmc0 all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb]
        # 28,695,763 ml1_pmc1     inprocess[sb] OR medline[sb] OR pubmed pmc[sb]
        assert a2n['all'] == a2n['all_ml0_pmc0'] + a2n['ml1_pmc1']
        #  3,116,339 pmnml_A              pubmednotmedline[sb]
        #  3,116,339 pmnml_B              pubmednotmedline[sb] NOT medline[sb]
        #  3,116,339 pmnml_C_ip0          pubmednotmedline[sb] NOT inprocess[sb]
        #          0 pmnml_0_ip1          pubmednotmedline[sb] AND inprocess[sb]
        #  1,612,274 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
        #  1,504,065 pmnml_A_pmc0         pubmednotmedline[sb] NOT pubmed pmc[sb]
        assert a2n['pmnml_A'] == a2n['pmnml_A_pmc1'] + a2n['pmnml_A_pmc0']
        #  5,323,939 pmc_all              pubmed pmc[sb]
        #    135,287 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
        #  3,503,057 medline_pmc1         medline[sb] AND pubmed pmc[sb]
        #  1,612,274 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
        pmc_notmedline = a2n['pmc_all'] - a2n['medline_pmc1'] - a2n['inprocess_A_pmc1']
        print('  {N:10,} of {M:10,} PMC (not MEDLINE)'.format(M=a2n['pmc_all'], N=pmc_notmedline))
        print(pmc_notmedline - a2n['pmnml_A_pmc1'])
        print('  {N:10,} of {M:10,} PubMed(not MEDLINE, PMC)'.format(
            N=a2n['all_ml0_pmc0'], M=a2n['all']))
        print('  {T:10,} = {A:10,} (MEDLINE OR PMC) + {B:10,} (not MEDLINE OR PMC)'.format(
            T=a2n['ml1_pmc1'] + a2n['all_ml0_pmc0'],
            A=a2n['ml1_pmc1'],
            B=a2n['all_ml0_pmc0']))
        total = a2n['all']
        au_all = a2n['au_all']
        print('  {N:10,} of {M:10,} {P:3.5f}% author ms'.format(N=au_all, M=total, P=100.0*au_all/total))
        print('  {T:10,} = {A:10,} (MEDLINE OR PMC) + {B:10,} (not MEDLINE OR PMC)'.format(
            T=a2n['ml1_pmc1'] + a2n['all_ml0_pmc0'],
            A=a2n['ml1_pmc1'],
            B=a2n['all_ml0_pmc0']))

    def dnld_count(self, query):
        """Searches an NCBI database for a user search term, returns NCBI IDs."""
        dct = self.run_eutilscmd('esearch', db='pubmed', term=query, rettype='count', retmode='json')
        if 'count' in dct:
            return int(dct['count'])
        return dct


# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Read a file created by pmidcite and write simple text file of PMIDs"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from os.path import exists
import argparse

from pmidcite.cfg import get_cfgparser

## from pmidcite.eutils.cmds.pubmed import PubMed
from pmidcite.cli.entry_keyset import get_details_cites_refs
## from pmidcite.cli.utils import get_mode_force
from pmidcite.cli.utils import get_outfile
from pmidcite.cli.utils import mk_outname_icite
## from pmidcite.cli.utils import wr_pmids
from pmidcite.cli.utils import read_pmids
## from pmidcite.cli.utils import read_top_pmids

from pmidcite.icite.nih_grouper import NihGrouper
from pmidcite.icite.downloader import get_downloader


class ReadPmids:
    """Read a file created by pmidcite and write simple text file of PMIDs"""

    def __init__(self):
        self.cfg = get_cfgparser()

    def _get_argparser(self):
        """Argument parser for Python wrapper of NIH's iCite given PubMed IDs"""
        parser = argparse.ArgumentParser(description="Run NIH's iCite given PubMed IDs")
        parser.add_argument(
            'infile', nargs='*',
            help='Read PMIDs from a pmidcite output file.')
        self.cfg.get_nihgrouper().add_arguments(parser)
        parser.add_argument(
            '-f', '--force_write', action='store_true',
            help='if an existing outfile file exists, overwrite it.')
        parser.add_argument(
            '-D', '--force_download', action='store_true',
            help='Download PMID iCite information to a Python file, over-writing if necessary.')
        parser.add_argument(
            '-o', '--outfile',
            help='Write current citation report to an ASCII text file.')
        parser.add_argument(
            '-c', '--load_citations', action='store_true', default=False,
            help='Load and print the list of citations for each paper.')
        parser.add_argument(
            '-r', '--load_references', action='store_true', default=False,
            help='Load and print the list of references for each paper.')
        parser.add_argument(
            '-R', '--no_references', action='store_true',
            help='Print the list of citations, but not the list of references.')
        return parser

    def cli(self):
        """Run iCite/PubMed using command-line interface"""
        argparser = self._get_argparser()
        args = argparser.parse_args()
        ## print('ARGS: ', args)
        pmids = self._read_pmids(args, argparser)
        # keys: outfile mode force_write
        ## dct = get_outfile(args.outfile, append_outfile, args.force_write)
        dct = get_outfile(args.outfile, None, args.force_write)
        outfile = self._get_outfile(args, dct)
        if pmids:
            self._wr_icite(outfile, pmids, args)
        else:
            print('XXX', outfile, pmids)

    def _wr_icite(self, outfile, pmids, args):
        """Write an iCite report for the provided PMIDs"""
        grouperobj = NihGrouper(args.min1, args.min2, args.min3, args.min4)
        details_cites_refs = get_details_cites_refs(
            args.verbose,
            args.load_citations,
            args.load_references,
            args.no_references)
        dnldr = get_downloader(
            details_cites_refs,
            grouperobj,
            self.cfg.get_dir_icite_py(),
            args.force_download)
        pmid2icitepaper_all = dnldr.get_pmid2paper(pmids, None)
        pmid2icitepaper_cur = {p: o for p, o in pmid2icitepaper_all.items() if o is not None}
        if outfile is not None:
            dnldr.wr_papers(outfile, pmid2icitepaper_cur, args.force_write, 'w')

    def _read_pmids(self, args, argparser):
        """Read PMIDs from a file"""
        ##print('ICITE ARGS: ../pmidcite/src/pmidcite/cli/icite.py', args)
        # Get user-specified PMIDs
        pmids = self.rd_pmidtxts(args.infile)
        if not pmids:
            argparser.print_help()
            print('**WARNING: NO PMIDs FOUND')
            return pmids
        return pmids

    @staticmethod
    def _get_outfile(args, dct):
        """Return outfile"""
        if dct['outfile']:
            return dct['outfile']
        if len(args.infile) == 1:
            return mk_outname_icite(args.infile[0])
        return None

    @staticmethod
    def rd_pmidtxts(fins_pmidcite):
        """Get PMIDs from the command line or from a file"""
        pmids = []
        seen = set()
        if fins_pmidcite:
            for fin in fins_pmidcite:
                if exists(fin):
                    ## TBD: for pmid in read_top_pmids(fin):
                    for pmid in read_pmids(fin):
                        if pmid not in seen:
                            pmids.append(pmid)
                            seen.add(pmid)
                else:
                    print('  MISSING: {FILE}'.format(FILE=fin))
                if len(fins_pmidcite) != 1:
                    print('{N:6} PMIDs READ'.format(N=len(pmids)))
        return pmids


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Plot the types of content and their amount in PubMed"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
from pmidcite.cfg import Cfg
from pmidcite.eutils.pubmed.counts.dnld import PubMedDnld
from pmidcite.eutils.pubmed.counts.plt import PubMedPlot


def main(dnld=False):
    """Plot the types of content and their amount in PubMed"""
    ## fpat_png = 'log/pubmed_content/pubmed_content_{DATE}.png'
    fout_pngs = [
        'pubmed_content_2020_01_10.png',
        'pubmed_content_2020_01_10.tiff',
        'pubmed_content_2020_01_10.pdf',
        'pubmed_content_2020_01_10.pdf',
    ]
    fout_py = 'src/pmidcite/eutils/pubmed/counts/dnlded_data.py'

    cfg = Cfg()
    obj = PubMedDnld(cfg.get_email(), cfg.get_apikey(), cfg.get_tool())
    name2cnt, date = obj.get_content_counts(fout_py, dnld)
    # fout_png = fpat_png.format(DATE=date)
    obj.prt_content_counts(name2cnt)
    print('    # {DATE}'.format(DATE=date))

    # Plot PubMed content
    plt = PubMedPlot(name2cnt)
    for fout_png in fout_pngs:
        plt.plt_content_counts(fout_png, dpi=800)
    for name, ntd in plt.dataobj.pltdata_pubmed.items():
        print(' {N:10,} {P:5.1f}% {NAME}'.format(NAME=name, N=ntd.count, P=ntd.perc))


if __name__ == '__main__':
    main(len(sys.argv) != 1)

# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Given PubMed IDs (PMIDs), download NIH citation data and write it to a Python module"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

## from timeit import default_timer
from collections import OrderedDict

import traceback
import requests

from pmidcite.icite.utils import split_list
## from tests.prt_hms import prt_hms


class NIHiCiteAPI:
    """Given a PubMed ID (PMID), return a list of publications which cite it from NIH's iCite"""

    opt_keys = {
        # Number of publications to return. The maximum allowed is 1000.
        'limit'
        # Only return publications with a PMID greater than this
        # Example: /api/pubs?offset=23456789&limit=10&format=csv
        'offset'
        # Only return publications from the given year.
        'year'
        # Only return publications with the given PubMed IDs.
        # Separate multiple IDs with commas to request up to 1000 at a time.
        # If this parameter is provided, all other parameters are ignored.
        'pmids'
        # only return publications with the given fields.
        # Separate multiple fields with commas (no space).
        # Field names are very specific and listed in Response example below.
        # No fl param will return all fields.
        # Example: /api/pubs?pmids=28968381,28324054,23843509&fl=pmid,year,title,apt
        'fl'
        # return csv (comma separated value) by specifying format=csv rather than the default JSON.
        'format'}

    #           https://icite.od.nih.gov/api
    url_base = 'https://icite.od.nih.gov/api/pubs'

    flds_yes_no = {'is_research_article', 'is_clinical', 'provisional'}
    yes_no = {'Yes':True, 'No':False}

    def __init__(self, **kws):
        self.kws = {k:v for k, v in kws.items() if k in self.opt_keys}
        self.msgs = []

    def dnld_nihdict(self, pmid):
        """Download NIH citation data for one researcher-spedified PMID. Return a corrected json"""
        rsp_json = self._send_request('{URL}/{PMID}'.format(URL=self.url_base, PMID=pmid))
        return self._adjust_jsondct(rsp_json) if rsp_json else None

    def dnld_nihdicts(self, pmids):
        """Download a list of NIH citation data for given PMIDs"""
        return self._dnld_gtmax(pmids) if len(pmids) > 1000 else self._dnld_ltmax(pmids)

    def _dnld_gtmax(self, pmids):
        """Run iCite on given PubMed IDs"""
        nih_dicts_all = []
        max_limit = 1000
        pmid_list_all = pmids if isinstance(pmids, list) else list(pmids)
        num_total = len(pmids)
        # The NIH-OCC allows for a maximum of 1,000 PMIDs to be downloaded at once
        for pmid_list_cur in split_list(pmid_list_all, max_limit):
            nih_dicts_cur = self._dnld_ltmax(pmid_list_cur)
            if nih_dicts_cur:
                nih_dicts_all.extend(nih_dicts_cur)
            # pylint: disable=line-too-long
            print('NIH citation data downloaded: {N:,} of {P:,}'.format(N=len(nih_dicts_all), P=num_total))
        return nih_dicts_all

    def _dnld_ltmax(self, pmids):
        """Download NIH citation data using a request using their API"""
        ## tic = default_timer()
        req_nihocc = '{URL}?pmids={PMIDS}'.format(
            URL=self.url_base,
            PMIDS=','.join(str(p) for p in pmids))
        ## tic = prt_hms(tic, "Create request")
        # Note: rsp_json['data'] returned from NIH not in same order as requested
        rsp_json = self._send_request(req_nihocc)
        ## tic = prt_hms(tic, "Send request. Get response")
        if rsp_json is not None:
            # Adjust the jsons downloaded for NIH citation data
            nih_dicts = []
            s_adjust_jsondct = self._adjust_jsondct
            pmids_downloaded = set()
            for nih_json_dct in rsp_json['data']:
                pmids_downloaded.add(nih_json_dct['pmid'])
                nih_dicts.append(s_adjust_jsondct(nih_json_dct))
            ## tic = prt_hms(tic, "Adjust reponse")
            # Report PMIDs that did not have NIH citation data downloaded
            pmids_missing = set(pmids).difference(pmids_downloaded)
            if pmids_missing:
                self._warn_missing(pmids_missing)
            return nih_dicts
        self._warn_missing(pmids)
        return None

    @staticmethod
    def _warn_missing(pmids):
        """Warn that NIH citation data was not downloaded for pmids"""
        print("**WARNING: {N:,} NIH CITATION DATA NOT DOWNLOADED FOR PMIDs: {PMIDs}".format(
            N=len(pmids), PMIDs=' '.join(str(s) for s in sorted(pmids))))

    def _send_request(self, cmd):
        """Send the request to iCite"""
        try:
            rsp = requests.get(cmd)
            if rsp.status_code == 200:
                return rsp.json()
            self._prt_errmsg(self._err_msg(rsp))
            return None
        except requests.exceptions.ConnectionError as errobj:
            self._prt_errmsg('**ERROR: ConnectionError = {ERR}\n'.format(ERR=str(errobj)))
            return None
        except:
            traceback.print_exc()
            raise RuntimeError('**ERROR DOWNLOADING {CMD}'.format(CMD=cmd))

    def _prt_errmsg(self, errmsg):
        """Print the error and add the error to the list of API messages"""
        print(errmsg)
        self.msgs.append(errmsg)

    @staticmethod
    def _err_msg(rsp):
        """Get error message if an NIH iCite request failed"""
        ## print('1 RRRRRRRRRRRRRRRRRRRRRR', dir(rsp))
        ## print('2 RRRRRRRRRRRRRRRRRRRRRR', rsp.status_code)
        ## print('3 RRRRRRRRRRRRRRRRRRRRRR', rsp.reason)
        ## print('4 RRRRRRRRRRRRRRRRRRRRRR', rsp.content)
        ## print('5 RRRRRRRRRRRRRRRRRRRRRR', rsp.text)
        ## #print('3 RRRRRRRRRRRRRRRRRRRRRR', rsp.json())
        ## print('6 RRRRRRRRRRRRRRRRRRRRRR', rsp.url)
        ## if rsp.json() is not None:
        ##     txt =' '.join('{K}({V})'.format(K=k, V=v) for k, v in sorted(rsp.json().items()))
        return '{CODE} {REASON} URL[{N}]: {URL}'.format(
            CODE=rsp.status_code,
            REASON=rsp.reason,
            N=len(rsp.url),
            URL=rsp.url)
            #TEXT=rsp.text)

    def _adjust_jsondct(self, json_dct):
        """Adjust values in the json dict"""
        dct = {}
        if json_dct['title'] is not None:
            title = json_dct['title'].strip()
            if '"' in title:
                title = title.replace('"', "'")
            if "\n" in title:
                title = title.replace('\n', " ")
            dct['title'] = title
        if json_dct['authors'] is not None:
            authors = json_dct['authors']
            if authors == '':
                dct['authors'] = []
            else:
                dct['authors'] = json_dct['authors'].split(', ')
        yes_no = self.yes_no
        dct['is_research_article'] = yes_no[json_dct['is_research_article']]
        dct['is_clinical'] = yes_no[json_dct['is_clinical']]
        dct['provisional'] = yes_no[json_dct['provisional']]
        lst = []
        lists = {'authors', 'cited_by_clin', 'cited_by', 'references'}
        for key, val in json_dct.items():
            if key in dct:
                lst.append((key, dct[key]))
            elif key in lists:
                lst.append((key, [] if val is None else val))
            else:
                lst.append((key, val))
        return OrderedDict(lst)

    @staticmethod
    def prt_dct(dct, prt):
        """Print NIH iCite data as a dict"""
        prt.write('"""Write data downloaded for NIH iCite data"""\n\n')
        prt.write('# pylint: disable=line-too-long\n')
        prt.write('ICITE = {\n')
        str_val = {'title', 'journal', 'doi'}
        for key, val in dct.items():
            if key == 'authors':
                prt.write("    '{K}': {V},\n".format(K=key, V=val))
                #prt.write("    '{K}': {AUTHORS},\n".format(
                #    K=key,
                #    AUTHORS='\n'.join(['"{AU}"'.format(AU=a) for a in val])))
            elif key not in str_val:
                prt.write("    '{K}': {V},\n".format(K=key, V=val))
            else:
                prt.write('''    '{K}': """{V}""",\n'''.format(K=key, V=val))
        prt.write('}\n')


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""This script downloads NCBI records based on user search query"""

# Entrez Help: # https://www.ncbi.nlm.nih.gov/books/NBK3837/
#
# A General Introduction to the E-utilities: # https://www.ncbi.nlm.nih.gov/books/NBK25497/
#
# Entrez-Utilities Help: # http://www.ncbi.nlm.nih.gov/books/NBK25501/
#
# The E-utilities In-Depth: Parameters, Syntax and More:
# http://www.ncbi.nlm.nih.gov/books/NBK25499/
# http://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch

# tags: # https://www.ncbi.nlm.nih.gov/books/NBK3827/table/pubmedhelp.Tn/


__copyright__ = "Copyright (C) 2014-present DV Klopfenstein. All rights reserved."
__author__ = 'DV Klopfenstein'

import sys
import traceback
import json
import collections as cx
import re
from xml.etree import ElementTree

import time

import urllib
from urllib.request import urlopen
from urllib.parse import urlencode
from urllib.error import URLError as _URLError
from urllib.error import HTTPError as _HTTPError


# pylint: disable=useless-object-inheritance
class EntrezUtilities(object):
    """Helper class for Entrez E-Utilities"""

    cgifmt = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/{ECMD}.fcgi"
    pat = 'PMIDs/epost={P} PMIDs/efetch={F} querykey({Q} of {Qmax}) start({S})'

    max_tries = 3
    sleep_between_tries = 15

    def __init__(self, email, apikey, tool, log=sys.stdout):
        self.email = email
        self.api_key = apikey
        self.tool = tool
        self.previous = 0
        self.log = log

    def get_database_list(self):
        """Get a list of Entrez databases"""
        # einfo: http://www.ncbi.nlm.nih.gov/books/NBK25499/
        return self.run_eutilscmd('einfo', retmode='json')

    def epost_ids(self, ids, database, num_ids_p_epost, retmax, **medline_text):
        """Post IDs using EPost"""
        epost_rsp = self.epost(database, ids, num_ids_p_epost=num_ids_p_epost)
        ## print('EPOST RSP', epost_rsp)
        # Set EFetch params
        efetch_params = dict(medline_text)
        efetch_params['webenv'] = epost_rsp['webenv']
        efetch_params['retmax'] = retmax  # num_ids_p_efetch
        # Run EFetches
        efetch_idxs = self._get_efetch_indices(epost_rsp, retmax, len(ids))
        #efetch_idxs = self._get_efetch_indices(epost_rsp, retmax, epost_rsp['count'])
        return efetch_idxs, efetch_params

    def _get_efetch_indices(self, epost_rsp, num_pmids_p_efetch, num_pmids):
        """Get EFetech list of: querykey_cur, pmids_cur, start"""
        # pmids num_pmids_p_epost num_pmids_p_efetch  ->  num_efetches
        # ----- ----------------- ------------------      ------------
        #     5                 2                  3                 3
        #     5                 2                  1                 5
        nts = []
        querykey_max = epost_rsp['querykey']
        num_pmids_p_epost_cur = epost_rsp['num_ids_p_epost']
        pat_val = {
            'P':epost_rsp['num_ids_p_epost'],
            'F':num_pmids_p_efetch,
            'Qmax':epost_rsp['querykey']}
        for querykey_cur, pmids_cur in enumerate(epost_rsp['qkey2ids'], 1):
            #### if querykey_cur == querykey_max:
            if querykey_cur == querykey_max and querykey_max != 1:
                num_pmids_p_epost_cur = num_pmids%epost_rsp['num_ids_p_epost']
            for start in range(0, num_pmids_p_epost_cur, num_pmids_p_efetch):
                desc = self.pat.format(Q=querykey_cur, S=start, **pat_val)
                # pylint: disable=line-too-long
                pmids_exp = pmids_cur[start:start+num_pmids_p_efetch] if pmids_cur is not None else None
                if pmids_exp:
                    nts.append([desc, start, pmids_exp, querykey_cur])
        return nts

    def _run_efetch(self, database, start, querykey, pmids_exp, desc, **params):
        """Get text from EFetch response"""
        rsp_dct = self.run_req('efetch', retstart=start, query_key=querykey, db=database, **params)
        if rsp_dct is None:
            print('\n{DESC}\n**ERROR: DATA is None'.format(DESC=desc))
            return None
        rsp_txt = rsp_dct['data'].decode('utf-8')
        err_txt = self._chk_error_str(rsp_txt)
        if err_txt is not None:
            print('\nURL: {URL}\n{DESC}\n**ERROR: {ERR}'.format(
                ERR=err_txt, DESC=desc, URL=rsp_dct['url']))
            return None
        ## # PubMed test
        ## pmids_downloaded = [int(i) for i in re.findall(r'PMID-\s*(\d+)', rsp_txt)]
        ## #print(desc, pmids_downloaded, pmids_exp)
        ## if pmids_exp:
        ##     # pylint: disable=line-too-long
        ##     assert pmids_downloaded == pmids_exp, '{TXT}\nDESC: {DESC}\nDL[{D}]: {DL}\nEXP[{L}]: {EXP}'.format(
        ##         TXT=rsp_txt, DESC=desc, D=len(pmids_downloaded),
        ##         DL=pmids_downloaded, EXP=pmids_exp, L=len(pmids_exp))
        return rsp_txt

    @staticmethod
    def _chk_error_str(text):
        """Check if data was correctly downloaded"""
        p0_err = text.find('<ERROR>')
        if p0_err < 0:
            return None
        msg = text[p0_err+7:]
        p1_err = msg.find('</ERROR>')
        return msg[:p1_err]

    #### def pubmed_query_fetch(self, query):
    ####     """Given a PubMed query, return results as  text"""
    ####     rsp_qry = self.find_ids_with_esearch('pubmed', query, retmax=500)
    ####     if rsp_qry is None:
    ####         return {'TEXT':None, 'RSP_QUERY': rsp_qry, 'RSP_POST':None}
    ####     rsp_post = self.epost('pubmed', rsp_qry['idlist'], num_ids_p_epost=10)
    ####     txt = self.run_eutilscmd(
    ####         'efetch',
    ####         db='pubmed',
    ####         retstart=0,
    ####         retmax=500,          # max: 10,000
    ####         rettype='medline',
    ####         retmode='text',
    ####         webenv=rsp_post['webenv'],
    ####         query_key=rsp_post['querykey'])
    ####     return {'TEXT':txt, 'RSP_QUERY': rsp_qry, 'RSP_POST':rsp_post}

    #### def find_ids_with_esearch(self, database, query, retmax, **esearch):
    ####     """Searches an NCBI database for a user search term, returns NCBI IDs."""
    ####     dct = self.run_eutilscmd(
    ####         'esearch',
    ####         db=database,
    ####         term=query,
    ####         retmax=retmax,
    ####         usehistory="y", # NCBI prefers we use history(QueryKey, WebEnv) for next acess
    ####         retmode='json',
    ####         **esearch)

    ####     if dct is not None and 'idlist' in dct and dct['idlist']:
    ####         # KEYS: count retmax retstart querykey webenv idlist
    ####         #       translationset translationstack querytranslation warninglist
    ####         dct['idlist'] = [int(n) for n in dct['idlist']]
    ####         dct['count'] = int(dct['count'])
    ####         # Note: If the maximum number of PMIDs are found, won't get the ones after
    ####         assert len(dct['idlist']) < retmax, 'EntrezUtilities: PMIDS({N}) > retmax({M})'.format(
    ####             M=len(dct['idlist']), N=dct['count'])
    ####         return dct
    ####     return None

    def epost(self, database, ids, num_ids_p_epost=10):
        """Posts to NCBI WebServer of any number of UIDs."""
        # Load the first 1...(step-1) UIDs to entrez-utils using epost. Get WebEnv to finish post
        if not isinstance(ids, list):
            raise RuntimeError('\n**FATAL: IDs({})\n**FATAL: EPost IDs NOT A LIST'.format(ids))
        if not ids:
            print('**NOTE: NO IDs to EPost')
            return None
        ret = {
            'num_ids_p_epost': num_ids_p_epost,
            'qkey2ids': [ids[:num_ids_p_epost]]
        }
        str_ids = [str(n) for n in ids]
        id_str = ','.join(str_ids[:num_ids_p_epost])
        # epost produces WebEnv value ($web1) and QueryKey value ($key1)
        rsp = self.run_eutilscmd('epost', db=database, id=id_str)
        if 'webenv' in rsp:
            if self.log is not None:
                ## self.log.write('FIRST EPOST RESULT: {}\n'.format(rsp))
                self.log.write('epost webenv: {W}\n'.format(W=rsp['webenv']))
                self.log.write("epost querykey({Q:>6}) ids[{N}]={Ps}\n".format(
                    N=len(str_ids), Q=rsp['querykey'], Ps=id_str))
            ret['webenv'] = rsp['webenv']
            webenv = rsp['webenv']
            num_ids = len(ids)
            # Load the remainder of the UIDs using epost
            for idx in range(num_ids_p_epost, num_ids, num_ids_p_epost):
                end_pt = idx+num_ids_p_epost
                if num_ids < end_pt:
                    end_pt = num_ids
                #print '{:3} {:3} {:3}'.format(num_ids, idx, end_pt)
                id_str = ','.join(str_ids[idx:end_pt])
                rsp = self.run_eutilscmd('epost', db=database, id=id_str, webenv=webenv)
                ret['qkey2ids'].append(ids[idx:idx+num_ids_p_epost])
                webenv = rsp['webenv']
                if self.log is not None:
                    self.log.write("epost querykey({Q:>6}) ids[{N}]={Ps}\n".format(
                        N=end_pt-idx, Q=rsp['querykey'], Ps=id_str))
        elif 'error' in rsp:
            raise RuntimeError('**ERROR EPost: {MSG}'.format(MSG=rsp['error']))
        else:
            print(rsp)
            raise Exception("NO webenv RETURNED FROM FIRST EPOST")
        ## if self.log is not None:
        ##     self.log.write('LAST  EPOST RESULT: {}\n'.format(rsp))
        ret['querykey'] = rsp['querykey']
        return ret

    @staticmethod
    def _return_einforesult(record):
        """Return EInfo result"""
        einforesult = record['einforesult']
        cmdtype = record['header']['type']
        if 'dblist' in einforesult:
            return einforesult['dblist']
        if cmdtype == 'einfo' and 'dbinfo' in einforesult:
            assert len(record['einforesult']['dbinfo']) == 1
            ## print('RRRRRRRRRRRRRRR', record.keys())
            ## print('RRRRRRRRRRRRRRR', len(record['einforesult']['dbinfo']))
            ## print('RRRRRRRRRRRRRRR', record)
            return record['einforesult']['dbinfo'][0]
        raise RuntimeError('IMPLEMENT _return_einforesult')

    @staticmethod
    def _return_linksets(record):
        """Return ELink result"""
        links_all = []
        for dct0 in record['linksets']:
            ## print('DCT', dct0)
            if 'linksetdbs' in dct0:
                for dct1 in dct0['linksetdbs']:
                    links_all.extend(dct1['links'])
        print('{N} LINKED ITEMS'.format(N=len(links_all)))
        return links_all

    # ------------------------------------------------------------------------------------
    def run_eutilscmd(self, cmd, **params):  # params=None, post=None, ecitmatch=False):
        """Run NCBI E-Utilities command"""
        # params example: db retstart retmax rettype retmode webenv query_key
        rsp_dct = self.run_req(cmd, **params) # post=None, ecitmatch=False):
        ## print('RRRRRRRRRRRRRRRRRRRRRRR', rsp_dct)
        if rsp_dct is not None:
            return self._extract_rsp(rsp_dct['data'], params.get('retmode'))
        return None

    def _mk_cgi(self, cmd, **params):
        """Get Fast Common Gateway Interface (fcgi) string, given E-utils command/parameters"""
        cgi = self.cgifmt.format(ECMD=cmd)
        params = self._construct_params(params)
        ## print('PARAMS', params)
        options = self._encode_options(params)
        cgi += '?' + options
        return cgi

    def run_req(self, cmd, prt=None, **params):  # params=None, post=None, ecitmatch=False):
        """Run NCBI E-Utilities command"""
        # params example: db retstart retmax rettype retmode webenv query_key
        cgi = self._mk_cgi(cmd, **params)
        ## print('CGI: {CGI}\n'.format(CGI=cgi))
        try:
            rsp = self._run_req(cgi, prt) # post=None, ecitmatch=False):
            # print('PPPPPPPPPPPPPPPPPP src/pmidcite/eutils/cmds/base.py run_req', rsp)
            return rsp
        except json.decoder.JSONDecodeError as errobj:
            print('\n**FATAL: CGI: {CGI}'.format(CGI=cgi))
            print('**FATAL: JSONDecodeError = {ERR}\n'.format(ERR=str(errobj)))
            traceback.print_exc()
        except urllib.error.HTTPError as errobj:
            print('\n**FATAL: CGI: {CGI}'.format(CGI=cgi))
            print('**FATAL: {ERR}\n'.format(ERR=str(errobj)))
            traceback.print_exc()
        except urllib.error.ContentTooShortError as errobj:
            print('\n**FATAL: CGI: {CGI}'.format(CGI=cgi))
            print('**FATAL: ContentTooShortError = {ERR}\n'.format(ERR=str(errobj.reason)))
            traceback.print_exc()
        except urllib.error.URLError as errobj:
            print('\n**FATAL: CGI: {CGI}'.format(CGI=cgi))
            print('**FATAL: URLError = {ERR}\n'.format(ERR=str(errobj.reason)))
            traceback.print_exc()
        except RuntimeError as errobj:
            print('\n**FATAL: CGI: {CGI}\n'.format(CGI=cgi))
            print(errobj)
            traceback.print_exc()

    def _run_req(self, cgi, prt):  # params=None, post=None, ecitmatch=False):
        """Get a response from running a Entrez Utilities command"""
        # NCBI requirement: At most three queries per second if no API key is provided.
        # Equivalently, at least a third of second between queries
        # Using just 0.333333334 seconds sometimes hit the NCBI rate limit,
        # the slightly longer pause of 0.37 seconds has been more reliable.
        delay = 0.1 if self.api_key else 0.37
        current = time.time()
        wait = self.previous + delay - current
        if wait > 0:
            time.sleep(wait)
            self.previous = current + wait
        else:
            self.previous = current
        if prt:
            prt.write('CMD: {CMD}\n'.format(CMD=cgi))
        ## sys.stdout.write('CMD: {CMD}\n'.format(CMD=cgi))  # For PubMed MISMATCHES

        for idx in range(self.max_tries):
            try:
                with urlopen(cgi) as rsp:
                    assert rsp.status == 200, dir(rsp)
                    return {
                        'code': rsp.getcode(),  # code/status
                        'msg': rsp.msg,
                        'url': rsp.geturl(),
                        'headers': cx.OrderedDict(rsp.getheaders()),
                        'data': rsp.read()}
            except _URLError as exception:
                # Reraise if the final try fails
                if idx >= self.max_tries - 1:
                    raise

                # Reraise if the exception is triggered by a HTTP 4XX error
                # indicating some kind of bad request
                if isinstance(exception, _HTTPError): ##? and exception.status // 100 == 4:
                    raise

                # Treat everything else as a transient error and try again after a
                # brief delay.
                time.sleep(self.sleep_between_tries)

    def _extract_rsp(self, record, retmode):
        """Extract the data from a response from running a Entrez Utilities command"""
        if retmode == 'json':
            try:
                dct = json.loads(record)
                if 'esearchresult' in dct:
                    return dct['esearchresult']
                if 'einforesult' in dct:
                    return self._return_einforesult(dct)
                if 'linksets' in dct:
                    return self._return_linksets(dct)
                print('KEYS:', dct.keys())
                print('DCT:', dct)
                raise RuntimeError('UNKNOWN RESULT in _run_req')
            except json.decoder.JSONDecodeError as errobj:
                print('JSONDecodeError = {ERR}'.format(ERR=str(errobj)))
                traceback.print_exc()
                print('\n**FATAL JSONDecodeError:\n{RECORD}'.format(RECORD=record.decode('utf-8')))

        if retmode == 'text':
            ## print('RECORD:', str(record))
            return record.decode('utf-8')

        ## print('RETMODE', retmode)
        ## print('RECORD', record)

        ## print(record)
        # <?xml version="1.0" encoding="ISO-8859-1"?>
        # <!DOCTYPE ePostResult
        #      SYSTEM "https://eutils.ncbi.nlm.nih.gov/eutils/dtd/20090526/epost.dtd"
        #      PUBLIC "-//NLM//DTD epost 20090526//EN">
        # <ePostResult>
        #     <QueryKey>1</QueryKey>
        #     <WebEnv>NCID_1_14223415_130.14.18.97 ... </WebEnv>
        # </ePostResult>
        # Parse XML
        root = ElementTree.fromstring(record)
        ## print('root.tag', root.tag)
        assert root.tag in 'ePostResult', root.tag
        dct = {r.tag.lower():r.text for r in root}
        if 'querykey' in dct:
            dct['querykey'] = int(dct['querykey'])
        ## print('XML:', root)
        ## print('XML:', root.tag)
        ## print('XML:', root.attrib)
        ## print('XML:', dir(root))
        ## print('XML[0]({}) ({})'.format(root[0].tag, root[0].text))
        ## print('XML[1]({})'.format(root[1].tag))
        ## print('XML:', dir(root[0]))
        ## print('XML:', dct)
        return dct

    def _construct_params(self, params):
        if params is None:
            params = {}

        # Remove None values from the parameters
        for key, value in list(params.items()):
            if value is None:
                del params[key]
        # Tell Entrez that we are using Biopython (or whatever the user has
        # specified explicitly in the parameters or by changing the default)
        if "tool" not in params:
            params["tool"] = self.tool
        # Tell Entrez who we are
        if "email" not in params:
            params["email"] = self.email
        if self.api_key and "api_key" not in params:
            params["api_key"] = self.api_key
        return params

    def _encode_options(self, params):
        """Encode request for url"""
        # If any values in the query arg are sequences and doseq is true, each
        # sequence element is converted to a separate parameter.
        doseq = self._get_doseq(params)
        ## print('DOSEQ', doseq)
        options = urlencode(params, doseq=doseq)
        ## print('ENCODED:', options)
        # urlencode encodes pipes, which NCBI expects in ECitMatch
        if 'ecitmatch' in params:
            options = options.replace("%7C", "|")
        return options

    @staticmethod
    def _get_doseq(params):
        """If doseq==True, each sequence element is converted to a separate parameter"""
        if 'term' in params:
            term = params['term']
            ## print('TERM', term)
            if '[Title]' in term or '[TI]' in term:
                return False
        return True

    @staticmethod
    def _get_num_iterations(count, retmax):
        """Get the number of iterations given the total and the max returned per time"""
        num_tot = count//retmax
        if count%retmax != 0:
            num_tot += 1
        return num_tot


# Copyright (C) 2014-present DV Klopfenstein. All rights reserved.



"""Given a PubMed ID (PMID), download a list of publications which cite and reference it"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from os.path import exists
from os.path import join

from pmidcite.icite.dnldr.pmid_dnlder_base import NIHiCiteDownloaderBase
from pmidcite.icite.dnldr.pmid_loader import NIHiCiteLoader
from pmidcite.icite.entry import NIHiCiteEntry


class NIHiCiteDownloader(NIHiCiteDownloaderBase):
    """Given a PubMed ID (PMID), download a list of publications which cite and reference it"""

    def __init__(self, dir_download, force_download, details_cites_refs=None, nih_grouper=None):
        super(NIHiCiteDownloader, self).__init__(details_cites_refs, nih_grouper)
        self.dnld_force = force_download
        self.dir_dnld = dir_download  # Recommended dir_icite_py: ./icite
        self.loader = NIHiCiteLoader(self.nihgrouper, dir_download, self.details_cites_refs)
        if not exists(dir_download):
            raise RuntimeError('**FATAL: NO DIRECTORY: {DIR}'.format(DIR=dir_download))

    def get_icites(self, pmids):
        """Download NIH iCite data for requested PMIDs"""
        # Python module filenames
        s_dir_dnld = self.dir_dnld
        pmid2py = {p:join(s_dir_dnld, 'p{PMID}.py'.format(PMID=p)) for p in pmids}
        if self.dnld_force:
            pmid2nihentry = {o.pmid: o for o in self._dnld_icites(pmid2py)}
            return [pmid2nihentry[pmid] for pmid in pmids if pmid in pmid2nihentry]
        # Separate PMIDs into those stored in Python modules and those not
        nihentries_all = []
        pmids_pyexist1 = set(pmid for pmid, py in pmid2py.items() if exists(py))
        pmids_pyexist0 = set(pmids).difference(pmids_pyexist1)
        if pmids_pyexist1:
            nihentries_loaded = self._load_icites(pmids_pyexist1, pmid2py)
            if nihentries_loaded:
                nihentries_all.extend(nihentries_loaded)
        if pmids_pyexist0:
            nihentries_all.extend(self._dnld_icites({p:pmid2py[p] for p in pmids_pyexist0}))
        # Return results sorted in the same order as input PMIDs
        pmid2nihentry = {o.pmid:o for o in nihentries_all}
        return [pmid2nihentry[pmid] for pmid in pmids if pmid in pmid2nihentry]

    def _dnld_icites(self, pmid2foutpy):
        """Download a list of NIH citation data for PMIDs"""
        nihdicts = self.api.dnld_nihdicts(pmid2foutpy.keys())
        if nihdicts:
            s_wrpy = self._wrpy
            for nih_dict in nihdicts:
                s_wrpy(pmid2foutpy[nih_dict['pmid']], nih_dict)
            s_get_group = self.nihgrouper.get_group
            return [NIHiCiteEntry(d, s_get_group(d['nih_percentile'])) for d in nihdicts]
        return []

    def get_icite(self, pmid):
        """Load or download NIH iCite data for requested PMID"""
        file_pmid = join(self.dir_dnld, 'p{PMID}.py'.format(PMID=pmid))
        if self.dnld_force or not exists(file_pmid):
            nih_dict = self.api.dnld_nihdict(pmid)
            if nih_dict:
                self._wrpy(file_pmid, nih_dict)
                return NIHiCiteEntry(
                    nih_dict,
                    self.nihgrouper.get_group(nih_dict['nih_percentile']))
        return self.loader.load_icite(file_pmid)  # NIHiCiteEntry

    # -------------------------------------------------------------------------------------
    def _get_pmids_missing(self, pmids_all):
        """Get PMIDs that have not yet been downloaded"""
        pmids_missing = set()
        for pmid_cur in pmids_all:
            file_pmid = join(self.dir_dnld, 'p{PMID}.py'.format(PMID=pmid_cur))
            if not exists(file_pmid):
                pmids_missing.add(pmid_cur)
        return pmids_missing

    def _wrpy(self, fout_py, dct, log=None):
        """Write NIH iCite to a Python module"""
        with open(fout_py, 'w') as prt:
            self.api.prt_dct(dct, prt)
            # Setting prt to sys.stdout -> WROTE: ./icite/p10802651.py
            if log:
                log.write('  WROTE: {PY}\n'.format(PY=fout_py))

    def _load_icites(self, pmids, pmid2py):
        """Load a list of NIH citation data for PMIDs"""
        nihentries_loaded = []
        s_load_icite = self.loader.load_icite
        num_exist = len(pmids)
        for idx, pmid in enumerate(pmids, 1):
            nihentries_loaded.append(s_load_icite(pmid2py[pmid]))
            if idx%1000 == 0:
                print('NIH citation data loaded: {N:,} of {M:,}'.format(N=idx, M=num_exist))
        ## nihentries_all.extend([s_load_icite(pmid2py[p]) for p in pmids_pyexist1])
        return nihentries_loaded


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Given a PubMed ID (PMID), return a list of citing publications"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from pmidcite.cli.icite import NIHiCiteCli  # get_argparser
from pmidcite.cfg import get_cfgparser


def main():
    """Print lists of pubs in formation"""
    NIHiCiteCli(get_cfgparser(prt=None)).cli()


if __name__ == '__main__':
    main()

# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Plot groups in the NIH percentile map showing how well a paper is doing among its peers"""

__copyright__ = "Copyright (C) 2020-present, DV Klopfenstein. All rights reserved."

import math
from sys import stdout

# pylint: disable=wrong-import-position
import matplotlib as mpl
mpl.use('agg')
## import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib import colors
import numpy as np
import scipy.stats as stats


class PltNihPercentile:
    """Plot groups in the NIH percentile map showing how well a paper is doing among its peers"""

    MU = 0
    VARIANCE = 1
    SIGMA = math.sqrt(VARIANCE)

    xvals = np.linspace(MU - 3*SIGMA, MU + 3*SIGMA, 120)
    yvals = stats.norm.pdf(xvals, MU, SIGMA)

    def plt_normal_curve(self, axes):
        """Plot the normal curve"""
        axes.plot(self.xvals, self.yvals, color='k')

    def get_vline_yvals(self, div_lines):
        """Get the y-value for the given vertical lines dividing percentile groups"""
        # https://stackoverflow.com/questions/20864847/probability-to-z-score-and-vice-versa
        # zscore-to-percentile: stats.norm.cdf(val)
        # percentile-to-zscore: stats.norm.pdf(val)
        return {xval:stats.norm.pdf(xval, self.MU, self.SIGMA) for xval in div_lines}

    @staticmethod
    def get_colors(colormap):
        """Get colors for each NIH percentile group"""
        cmap = cm.get_cmap(colormap)
        norm = colors.Normalize(vmin=0, vmax=4)
        scalarmap = cm.ScalarMappable(norm=norm, cmap=cmap)
        rgbas = [scalarmap.to_rgba(n) for n in range(5)]
        # NIH Group 0 color: # ff0029
        # NIH Group 1 color: # cdff00
        # NIH Group 2 color: # 00ff8c
        # NIH Group 3 color: # 0018ff
        # NIH Group 4 color: # ff00bf
        for idx, rgba in enumerate(rgbas):
            print('NIH Group {G} color: # {COL}'.format(
                G=idx,
                COL=str(mpl.colors.rgb2hex(rgba))[1:]))
        return rgbas

    def fill_between(self, axes, xval0, xval1, facecolor):
        """Color between standard deviation lines"""
        xvals = np.linspace(xval0, xval1, 20)
        yvals = stats.norm.pdf(xvals, self.MU, self.SIGMA)
        axes.fill_between(xvals, yvals, facecolor=facecolor, alpha=.4)


class PltNihVals(PltNihPercentile):
    """Plot groups in the NIH percentile map showing how well a paper is doing among its peers"""


    def __init__(self, group_mins):
        self.group_mins = group_mins  # List of dividers w/values from 0 to 1
        # https://stackoverflow.com/questions/20864847/probability-to-z-score-and-vice-versa
        self.zscores = [stats.norm.ppf(v/100) for v in group_mins]
        self.vlines_x = [-3] + self.zscores + [3]
        self.vlines_y = [stats.norm.pdf(x, self.MU, self.SIGMA) for x in self.vlines_x]

    def prt_vals(self, prt=stdout):
        """Print values and their z-scores"""
        for perc, zscore in zip(self.group_mins, self.zscores):
            prt.write('{PERC:5.3f} {Z}\n'.format(PERC=perc, Z=zscore))

    def get_group_percentages(self):
        """Get the percentage that each group represents"""
        return [self.group_mins[0],
                self.group_mins[1] - self.group_mins[0],
                self.group_mins[2] - self.group_mins[1],
                self.group_mins[3] - self.group_mins[2],
                100 - self.group_mins[3]]

    def plt_lines(self, axes):
        """Plot the lines bordering the normal distribution and the group dividers"""
        self.plt_normal_curve(axes)
        for xval, yval in zip(self.vlines_x, self.vlines_y):
            axes.vlines(xval, ymin=0, ymax=yval)
        axes.hlines(0, xmin=-3, xmax=3)

    def colorfill_groups(self, axes, colormap='gist_rainbow'):
        """Fill each nih percentile group with a color from the colormap"""
        color_list = self.get_colors(colormap)
        for idx, color in enumerate(color_list):
            self.fill_between(axes, self.vlines_x[idx], self.vlines_x[idx+1], color)

    def add_text_groups(self, axes):
        """Add text describing groups and content"""
        group_percs = self.get_group_percentages()
        kwtxt = {'ha':'center', 'va':'center', 'fontweight':'bold'}
        # 0) Lowest
        axes.text(-2.01, -.003, '0', fontsize=22, ha='right', va='bottom', fontweight='bold')
        # 1) Low
        axes.text(-1.24, .12, '1', fontsize=25, **kwtxt)
        axes.text(-1.5, .03, 'Low\n{:2.0f}%'.format(group_percs[1]), fontsize=15, **kwtxt)
        # 2) Good
        axes.text(0, .35, '2', fontsize=25, **kwtxt)
        axes.text(0, .03, 'Good\n{:2.0f}%'.format(group_percs[2]), fontsize=15, **kwtxt)
        # 3) Better
        axes.text(1.24, .12, '3', fontsize=25, **kwtxt)
        axes.text(1.5, .03, 'High\n{:2.0f}%'.format(group_percs[3]), fontsize=15, **kwtxt)
        # 4) Best
        axes.text(2.01, -.003, '4', fontsize=22, ha='left', va='bottom', fontweight='bold')

        axes.annotate('{:2.0f}%'.format(group_percs[0]), fontsize=15, xy=(2.5, 0.02),
                      xycoords='data', xytext=(2.7, .07),
                      arrowprops=dict(arrowstyle="->", color='black'), **kwtxt)
        axes.annotate('{:2.0f}%'.format(group_percs[4]), fontsize=15, xy=(-2.5, 0.02),
                      xycoords='data', xytext=(-2.7, .07),
                      arrowprops=dict(arrowstyle="->", color='black'), **kwtxt)

    def anno_groupmin(self, axes):
        """Add group_minN annotations"""
        kws = {
            'xycoords':'data',
            'fontsize':12,
            'va':'center',
            'arrowprops':dict(arrowstyle="->", color='black', linestyle='dotted'),
        }
        axes.annotate('group1_min={}'.format(self.group_mins[0]),
                      xytext=(-3, 0.30), xy=(self.vlines_x[1], self.vlines_y[1]), ha='left', **kws)

        axes.annotate('group2_min={}'.format(self.group_mins[1]),
                      xytext=(-3, 0.38), xy=(self.vlines_x[2], self.vlines_y[2]), ha='left', **kws)

        axes.annotate('group3_min={}'.format(self.group_mins[2]),
                      xytext=(3, 0.38), xy=(self.vlines_x[3], self.vlines_y[3]), ha='right', **kws)

        axes.annotate('group4_min={}'.format(self.group_mins[3]),
                      xytext=(3, 0.30), xy=(self.vlines_x[4], self.vlines_y[4]), ha='right', **kws)


# Copyright (C) 2020-present, DV Klopfenstein. All rights reserved.



"""Read a file created by pmidcite and write simple text file of PMIDs"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from os.path import exists
from os.path import split
from os.path import splitext
from os.path import join
from sys import stdout


def read_pmids(fin, prt=stdout):
    """Read PMIDs from a file. One PMID per line."""
    pmids = _str_to_int(_read_pmids(fin, top_cit_ref=None))
    if prt:
        prt.write('{N:6,} PMIDs READ: {FILE}\n'.format(N=len(pmids), FILE=fin))
    return pmids

def get_pmids(pmid_list, fin_pmids):
    """Get PMIDs from the command line or from a file"""
    return _str_to_int(get_all(pmid_list, fin_pmids))

def _str_to_int(pmids):
    """Convert a list of string PMIDs to integer PMIDs"""
    int_pmids = []
    for pmidstr in pmids:
        if isinstance(pmidstr, int):
            int_pmids.append(pmidstr)
        elif pmidstr.isdigit():
            int_pmids.append(int(pmidstr))
    return int_pmids

def get_all(pmid_list, fin_pmids, top_cit_ref=None):
    """Get PMIDs from the command line or from a file"""
    if not pmid_list and not fin_pmids:
        return []
    pmids = list(pmid_list)
    # Search PMIDs listed in files
    seen = set(pmids)
    if fin_pmids:
        for fin in fin_pmids:
            if exists(fin):
                for pmid in _read_pmids(fin, top_cit_ref):
                    if pmid not in seen:
                        pmids.append(pmid)
            else:
                print('  MISSING: {FILE}'.format(FILE=fin))
    return pmids

def _read_pmids(fin, top_cit_ref):
    """Read PMIDs from a file. One PMID per line."""
    pmids = []
    if top_cit_ref is None:
        top_cit_ref = {'TOP',}  # TOP, CIT, CLI, REF
    with open(fin) as ifstrm:
        for line in ifstrm:
            line = line.strip()
            if line[:1] == '#' or line == '':
                continue
            if line[:3] in top_cit_ref:
                pmid = pmid[4:].split(maxsplit=1)[0]
                if pmid.isdigit():
                    pmids.append(pmid)
            elif line.isdigit():
                pmids.append(line)
    return pmids

def read_top_pmids(pmidcite_txt, topset=None):
    """Get PMIDs already found in pmidcite.txt"""
    pmids = set()
    if topset is None:
        topset = {'TOP',}
    with open(pmidcite_txt) as ifstrm:
        for line in ifstrm:
            if line[:3] in topset:
                flds = line.split()
                if flds[1].isdigit():
                    pmids.add(int(flds[1]))
    return pmids

def wr_pmids(fout_txt, pmids, mode='w', log=stdout):
    """Write PMIDs into a text file, one PMID per line"""
    with open(fout_txt, mode) as prt:
        for pmid in pmids:
            prt.write('{PMID}\n'.format(PMID=pmid))
        if log:
            log.write('{N:6,} WROTE: {FOUT}\n'.format(N=len(pmids), FOUT=fout_txt))

def mk_outname_pmids(fin):
    """Given input file, return output file for PMIDs"""
    outdir, outname = split(fin)
    basename, ext = splitext(outname)
    return join(outdir, '{BASE}_pmids{EXT}'.format(BASE=basename, EXT=ext))

def mk_outname_icite(fin):
    """Given input file, return output file for PMIDs"""
    outdir, outname = split(fin)
    basename, ext = splitext(outname)
    return join(outdir, '{BASE}_icite{EXT}'.format(BASE=basename, EXT=ext))

def get_outfile(outfile, append_outfile, force_write):
    """Given arguments, return outfile"""
    mode, force_wr = _get_mode_force(outfile, force_write, append_outfile)
    return {
        'outfile': _get_outfile_resolved(outfile, append_outfile),
        'mode':mode,
        'force_write':force_wr}

def _get_mode_force(outfile, force_write, append_outfile):
    """Given arguments, return outfile"""
    # if '-o', only over-write existing file if explicitly requested
    if outfile is not None:
        return 'w', force_write
    # if '-a', always append given file or create the file if needed
    if append_outfile is not None:
        return 'a', True
    return 'w', True

def _get_outfile_resolved(outfile, append_outfile):
    """Given outfile (write_outfile) and append_outfile, return one resolved outfile name"""
    if outfile is not None:
        return outfile
    if append_outfile is not None:
        return append_outfile
    return None


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Configuration file for PMID Cite"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
from os.path import exists


PMIDCITERC = \
"""# Link citation data from the National Institute of Health (NIH) to PubMed IDs.

# The file .pmidciterc is a configuration file for the project, pmidcite, and its
# executables, which includes icite.

# To protect your security, store this file (.pmidciterc) in a directory that is
# not managed by git or other version management tools. If the .pmidciterc is in
# the home directory, set the environmental variable, PMIDCITECONF, so pmidcite
# scripts can read it:
#
#   $ export PMIDCITECONF=~/.pmidciterc
#
[pmidcite]

# --------------------------------------------------------------------------------
# When comparing numerous papers, use the group variables to place
# the most recent papers first, the highest rated papers next, and
# the lowest-rated papers at the end of the list of papers and citation data.
group1_min = 2.1
group2_min = 15.7
group3_min = 83.9
group4_min = 97.5

# --------------------------------------------------------------------------------
# NCBI E-Utils needs: email, apikey, and tool to download PubMed records.
# These fields are needed if:
#   * You plan to run a search in PubMed and download the resulting list of PMIDs
#   * You plan to download abstracts from NCBI
#
# Note: The fields are not needed for viewing NIH citation data for a given PMID
#
# Learn more about registering for a NCBI API key here:
# https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/new-api-keys-for-the-e-utilities
email = name@university.edu
apikey = LONG_HEX_NCBI_API_KEY
tool = scripts

# --------------------------------------------------------------------------------
# The default is to work on-line (dir_icite_py = None);
# NIH citations will always be downloaded.
#
# To work off-line and store citation data downloaded from the NIH,
# set `dir_icite_py' to a directory name. Each citation downloaded
# from the NIH will then be stored in its own p<PMID>.py file for
# later use when there is no internet available.
#
# Specify the directory to store citation data from the NIH that is downloaded
# using commands such as icite:
#
#  # PMID 33031632 is a new (-1) research (R) paper discussing humans (H)
#  # with 2 citations and 18 references that is authored by 2 (au[02]) people with
#  # the first author being DV Klopfenstein.
#  #
#  $ icite 33031632
#  TOP 33031632 R. H....  -1 i 2021     2  0  18 au[02](D V Klopfenstein) Commentary ...
#
# Thousands of NIH citation files may be downloaded for a literature search
# on a single subject. It is not necessary to save the NIH citation data and I
# recommend not saving these files using git managed from the cloud.
#
# Note: Add an __init__.py file to the ./icite directory
#       Add /icite to the .gitignore to avoid saving thousands of unecessary files using git
#
# Recommended value, None or ./icite:
# dir_icite_py = None
# dir_icite_py = ./icite  # $ mkdir ./icite
dir_icite_py = None

# --------------------------------------------------------------------------------
# Store abstracts and publication data downloaded form PubMed in a dedicated directory
#
# Recommended value (mkdir ./log; mkdir ./log/pubmed):
# dir_pubmed_txt = ./log/pubmed
dir_pubmed_txt = None

# --------------------------------------------------------------------------------
# Store files containing longer lists of PMIDs into a dedicated directory
# specified using the variable, dir_pmids.
# Files containing lists of PMIDs can be downloaded either from the PubMed GUI
# or by running a PubMed query from your copy of pmidcite's dnld_pmids.py script.
#
# Store NIH citation data for all PMIDs in a single file into a directory
# dedicated to citation data files using the variable, dir_icite.
#
# Recommended values (mkdir ./log; mkdir ./log/pmids; mkdir ./log/icite)
# dir_pmids = ./log/pmids
# dir_icite = ./log/icite
dir_pmids = None
dir_icite = None
"""

def prt_rcfile(prt=stdout):
    """Print a default PMIR icite configuration file"""
    prt.write('\n# == CONFIGURATION FILE BEGIN: .pmidciterc ====================================\n')
    prt.write(PMIDCITERC)
    prt.write('# == CONFIGURATION FILE END: .pmidciterc ======================================\n\n')

def wr_rcfile(cfgfile, force=False):
    """Write a sample configuration with default values set"""
    if not exists(cfgfile) or force:
        with open(cfgfile, 'w') as prt:
            prt.write(PMIDCITERC)
            print('  WROTE: {CFG}'.format(CFG=cfgfile))
            return True
    print('  EXISTS: {CFG} OVERWRITE WITH wr_rc(force=True)'.format(CFG=cfgfile))
    return False


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""For installing pmidcite package and accompanying scripts."""

from os.path import abspath
from os.path import dirname
from os.path import join
# from distutils.core import setup
from glob import glob
from setuptools import setup
# import versioneer


NAME = 'pmidcite'

PACKAGES = [
    'pmidcite',
    'pmidcite.cli',
    'pmidcite.eutils',
    'pmidcite.eutils.cmds',
    'pmidcite.eutils.pubmed',
    'pmidcite.eutils.pubmed.counts',
    'pmidcite.icite',
    'pmidcite.icite.dnldr',
    'pmidcite.plot',
]

PACKAGE_DIRS = {p:join('src', *p.split('.')) for p in PACKAGES}

def get_long_description():
    """Return the contents of the README.md as a string"""
    dir_cur = abspath(dirname(__file__))
    with open(join(dir_cur, 'README.md'), encoding='utf-8') as ifstrm:
        return ifstrm.read()


setup(
    name=NAME,
    ## version=versioneer.get_version(),
    version='0.0.38',
    author='DV Klopfenstein, PhD',
    author_email='dvklopfenstein@protonmail.com',
    ## cmdclass=versioneer.get_cmdclass(),
    packages=PACKAGES,
    package_dir=PACKAGE_DIRS,
    # https://stackoverflow.com/questions/18787036/difference-between-entry-points-console-scripts-and-scripts-in-setup-py
    # https://stackoverflow.com/questions/45114076/python-setuptools-using-scripts-keyword-in-setup-py
    # https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/#scripts
    scripts=glob('src/bin/*.py'),
    entry_points={
        'console_scripts':[
            'icite=src.bin.icite:main',
        ],
    },
    # https://pypi.org/classifiers/
    classifiers=[
        'Programming Language :: Python',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    url='http://github.com/dvklopfenstein/pmidcite',
    description="Augment's a PubMed literature search with citation data from NIH-OCC's iCite.",
    # https://packaging.python.org/guides/making-a-pypi-friendly-readme/
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    # install_requires=['docopt'],
)



"""Text query finds database UIDs for later use in ESummary, EFetch or ELink"""
# https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch

__copyright__ = "Copyright (C) 2020-present DV Klopfenstein. All rights reserved."
__author__ = 'DV Klopfenstein'

import sys
from pmidcite.eutils.cmds.base import EntrezUtilities


class QueryIDs(EntrezUtilities):
    """Text query finds database UIDs for later use in ESummary, EFetch or ELink"""

    # https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch
    exp_params = {
        # Required Parameters:
        'db',
        'term',
        # Optional Parameters - History Server
        'usehistory',
        'webenv',
        'querykey',
        # Optional Parameters – Retrieval
        'retstart',
        'retmax',
        'rettype',
        'retmode',
        'sort',
        'field',
        'idtype',
        # Optional Parameters – Dates
        'datetype',
        'reldate',
        'mindate',
        'maxdate',
    }

    def __init__(self, email, apikey, tool, prt=sys.stdout):
        super(QueryIDs, self).__init__(email, apikey, tool, prt)

    def dnld_query_ids(self, query, database, num_ids_p_epost=10):
        """Searches a NCBI database for a user query, writes resulting entries into one file."""
        rsp_dct = self.get_query_rsp(query, database, num_ids_p_epost)
        return self.get_ids(rsp_dct, query, database, num_ids_p_epost)

    def get_query_rsp(self, query, database, num_ids_p_epost=10):
        """Searches a NCBI database for a user query, writes resulting entries into one file."""
        # 1) Query PubMed/Protein/etc. Get first N (num_ids_p_epost) of the total PMIDs
        rsp_dct = self.query(database, query, retmax=num_ids_p_epost)
        if rsp_dct is None:
            if self.log:
                self.log.write('No {DB} entries found: {Q}\n'.format(DB=database, Q=query))
                self.log.flush()
            return []
        if rsp_dct and self.log:
            self.log.write('{N:6,} IDs FOR {DB} QUERY({Q})\n'.format(
                DB=database, N=rsp_dct['count'], Q=query))
            self.log.flush()
        return rsp_dct

    def get_ids(self, rsp_dct, query, database, num_ids_p_epost=10):
        """Download PMIDs, N (num_ids_p_epost) at a time"""
        ## print('WWWWWWWWWWWWWWWWWWWWW pmidcite/eutils/cmds/query_ids.py', rsp_dct)
        if not rsp_dct:
            return []
        ids = list(rsp_dct['idlist'])
        kws_p = {
            'webenv': rsp_dct['webenv'],
            'querykey': rsp_dct['querykey'],
            'retmax': num_ids_p_epost,
        }
        tot_ids = rsp_dct['count']
        ## print('WWWWWWWWWWWWWWWWWWWWWWWW', kws_p)
        for retnum in range(1, self._get_num_querykeys(num_ids_p_epost, tot_ids)):
            ## print('WWWWWWWWWWWWWWWWWWWWWWWW retnum', retnum)
            rsp_dct = self.query(database, query, retstart=num_ids_p_epost*retnum, **kws_p)
            if rsp_dct:
                ## print('WWWWWWWWWWWWWWWWWWWWWWWW idlist', rsp_dct['idlist'])
                ids.extend(rsp_dct['idlist'])
        assert tot_ids == len(set(ids)), 'PMIDS EXP({E}) ACT({A})'.format(
            E=tot_ids, A=len(set(ids)))
        return ids

    @staticmethod
    def _get_num_querykeys(num_ids_p_epost, num_pmids):
        """Get the number of querykeys necessary to process all PMIDs"""
        num_querykeys = num_pmids//num_ids_p_epost
        if num_pmids%num_ids_p_epost != 0:
            num_querykeys += 1
        return num_querykeys

    def query(self, database, query, **esearch):
        """Text query finds database UIDs for later use in ESummary, EFetch or ELink"""
        kws_exp = self.exp_params.difference({'db', 'term', 'rettype', 'usehistory', 'retmode'})
        kws_act = {k:v for k, v in esearch.items() if k in kws_exp}
        # Returns:
        #    count
        #    retmax
        #    retstart
        #    querykey
        #    webenv
        #    idlist
        #    translationset
        #    translationstack
        #    querytranslation
        dct = self.run_eutilscmd(
            'esearch',
            db=database,
            term=query,
            rettype='uilist',
            # retmax=esearch.get('retmax', 10),
            usehistory="y", # NCBI prefers we use history(QueryKey, WebEnv) for next acess
            retmode='json',
            **kws_act)
        if dct is not None and 'idlist' in dct and dct['idlist']:
            if database in {'pubmed',}:
                dct['idlist'] = [int(n) for n in dct['idlist']]
            for fldname in {'count', 'retmax'}:
                dct[fldname] = int(dct[fldname])
            return dct
        return None


# Copyright (C) 2020-present, DV Klopfenstein. All rights reserved.



"""Holds NIH iCite data for one PubMed ID (PMID)"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout


def sortby_year(obj):
    """Sort lists of iCite items"""
    return [-1*obj.dct['year'], -1*obj.dct['nih_percentile']]

def sortby_cite(obj):
    """Sort lists of iCite items"""
    return [-1*obj.dct['citation_count'], -1*obj.dct['year']]

def sortby_nih_group(obj):
    """Sort lists of iCite items"""
    dct = obj.dct
    return [-1*dct['nih_group'], -1*dct['year'], -1*dct['nih_perc'],
            -1*dct['citation_count'] + -1*dct['num_clin'],
            -1*dct['num_refs'],
            -1*dct['pmid']]


# pylint: disable=too-many-instance-attributes
class NIHiCitePaper:
    """Holds NIH iCite data for one PubMed ID (PMID)"""

    sortby_dct = {
        'nih_group': sortby_nih_group,
        'cite': sortby_cite,
        'year': sortby_year,
    }

    def __init__(self, pmid, pmid2icite, header=None, pmid2note=None):
        self.pmid = pmid
        self.pmid2icite = pmid2icite
        ## print('PPPPPPPP NIHiCitePaper({}, len(pmid2icite)={})'.format(pmid, len(pmid2icite)))
        self.hdr = header  # A header to print before a paper
        # A short pmid2note to print at end of cite line
        self.pmid2note = {} if pmid2note is None else pmid2note
        self.icite = pmid2icite.get(pmid)
        # Sets of NIHiCiteEntrys
        self.cited_by = self._init_pmids('cited_by')
        self.cited_by_clin = self._init_pmids('cited_by_clin')
        self.references = self._init_pmids('references')

    def get_icite(self):
        """Get NIHiCiteEntry for this paper"""
        return self.pmid2icite[self.pmid]

    def str_line(self):
        """Return a string summarizing the the paper described herein"""
        txt = str(self.icite)
        if not self.pmid2note:
            return txt
        if self.pmid in self.pmid2note:
            return '{TXT} {NOTE}'.format(TXT=txt, NOTE=self.pmid2note[self.pmid])
        return txt

    def prt_top(self, prt=stdout):
        """Print TOP paper"""
        prt.write('TOP {iCite}\n'.format(iCite=self.str_line()))

    @staticmethod
    def prt_keys(prt=stdout):
        """Print paper keys"""
        prt.write('    TOP: The paper requested by the researcher\n')
        prt.write('    CIT: A paper that cited TOP\n')
        prt.write('    CLI: A clinical paper that cited TOP\n')
        prt.write("    REF: A paper referenced in the TOP paper's bibliography\n")

    def prt_summary(self, prt=stdout, sortby_cites='nih_group', sortby_refs='nih_group'):
        """Print summary of paper"""
        if self.hdr:
            prt.write('NAME: {NAME}\n'.format(NAME=self.hdr))
        prt.write('TOP {iCite}\n'.format(iCite=self.str_line()))
        # Citations by clinical papers
        if self.cited_by_clin:
            prt.write('Cited by {N} Clinical papers:\n'.format(N=len(self.cited_by_clin)))
        self._prt_list(self.cited_by_clin, 'CLI', prt, sortby_cites)
        # Citations
        if self.cited_by:
            prt.write('{N} of {M} citations downloaded:\n'.format(
                N=len(self.cited_by),
                M=self.icite.dct['citation_count']))
        self._prt_list(self.cited_by, 'CIT', prt, sortby_cites)
        # References
        if self.references:
            prt.write('{N} of {M} References downloaded:\n'.format(
                N=len(self.references),
                M=self.icite.dct['num_refs']))
            self._prt_list(self.references, 'REF', prt, sortby_refs)

    def get_sorted(self, icites, sortby=None):
        """Get citations or references, sorted"""
        if sortby is None:
            return icites
        if sortby in self.sortby_dct:
            sortby = self.sortby_dct[sortby]
        return sorted(icites, key=sortby)

    def _prt_list(self, icites, desc, prt, sortby=None):
        """Print list of NIH iCites in summary format"""
        if sortby is not None:
            icites = self.get_sorted(icites, sortby)
        if self.pmid2note:
            s_pmid2note = self.pmid2note
            for icite in icites:
                if icite.pmid in s_pmid2note:
                    prt.write('{DESC} {iCite} {note}\n'.format(
                        DESC=desc, iCite=str(icite), note=s_pmid2note[icite.pmid]))
                else:
                    prt.write('{DESC} {iCite}\n'.format(DESC=desc, iCite=str(icite)))
            return
        for icite in icites:
            prt.write('{DESC} {iCite}\n'.format(DESC=desc, iCite=str(icite)))

    def _init_pmids(self, name):
        """Load citation/reference PMIDs, if the 'top' paper has NIH iCite data"""
        if self.icite is None:
            return []
        s_pmid2icite = self.pmid2icite
        return set(s_pmid2icite[pmid] for pmid in self.icite.dct[name] if pmid in s_pmid2icite)

    def __str__(self):
        """Get the line containing data downloaded from NIH iCite for only the featured paper"""
        return self.str_line()

    ## TBD:
    ## def __eq__(self, rhs):
    ##     if self.pmid != rhs.pmid:
    ##         return False
    ##     if self.cited_by != rhs.cited_by:
    ##         print('LHS:', self.cited_by)
    ##         print('RHS:', rhs.cited_by)
    ##         return False
    ##     return True


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Given PMIDs, download PubMed entries into a text file"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import argparse

import os
from pmidcite.eutils.cmds.pubmed import PubMed
from pmidcite.cfg import Cfg
from pmidcite.cli.utils import get_pmids


class DnldPubMed:
    """Given PMIDs, download PubMed entries into a text file"""

    def __init__(self):
        cfg = Cfg()
        self.cfgdct = cfg.cfgparser
        self.pubmed = PubMed(email=cfg.get_email(), apikey=cfg.get_apikey(), tool=cfg.get_tool())

    def get_argparser(self):
        """Argument parser for Python wrapper of NIH's iCite given PubMed IDs"""
        parser = argparse.ArgumentParser(description="Run NIH's iCite given PubMed IDs")
        dir_pubmed_txt = self.cfgdct['pmidcite']['dir_pubmed_txt']
        parser.add_argument(
            'pmids', metavar='PMID', type=int, nargs='*',
            help='PubMed IDs (PMIDs)')
        parser.add_argument(
            '-i', '--infile', nargs='*',
            help='Read PMIDs from a file containing one PMID per line.')
        ## parser.add_argument(
        ##     '-o', '--outfile',
        ##     help='Write current citation report to an ASCII text file.')
        ## parser.add_argument(
        ##     '-f', '--force_write', action='store_true',
        ##     help='if an existing outfile file exists, overwrite it.')
        parser.add_argument(
            '-D', '--force_download', action='store_true',
            help='Download PMID iCite information to a Python file, over-writing if necessary.')
        parser.add_argument(
            '--dir_pubmed_txt', default=dir_pubmed_txt,
            help='Write PubMed entry into directory (default={D})'.format(D=dir_pubmed_txt))
        ## parser.add_argument(
        ##     '-c', '--wordcloud_filename',
        ##     help='Output filename (i.e. pmids.png) for a word cloud plot for the given PMIDs')
        return parser

    def cli(self):
        """Run iCite/PubMed using command-line interface"""
        argparser = self.get_argparser()
        args = argparser.parse_args()
        # 1) Get PMIDs
        pmids = self._get_pmids(args, argparser)
        print('{N} PMIDs'.format(N=len(pmids)))
        # 2) Download PubMed entries.
        if pmids:
            self.pubmed.dnld_wr1_per_pmid(pmids, args.force_download, args.dir_pubmed_txt)

    @staticmethod
    def _get_pmids(args, argparser):
        """Extract PMIDs from required args and PMID files"""
        pmids = get_pmids(args.pmids, args.infile)
        if not pmids:
            argparser.print_help()
        return pmids


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Holds NIH iCite data for one PubMed ID (PMID)"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from collections import Counter
from collections import defaultdict


class NIHiCitePapers:
    """Holds NIH iCite data for one PubMed ID (PMID)"""

    def __init__(self, pmid2paper):
        self.pmid2paper = pmid2paper
        self.author2cnt = self._init_au2cnt()
        self.au2firstlast = self._init_au2firstlast(self.author2cnt.keys())

    def get_pmid2fl(self, authors):
        """Given a list of author names, return pmid2[FfL]"""
        # {'Jo Lynne Rokita', 'Jo Lynne Harenza', 'J L Harenza'})
        pmid2firstlast = {}
        for author in authors:
            if author in self.au2firstlast:
                ## print('FFFFFFFFFFFFFFFFFFF', author, firstlast)
                for mark, pmids in self.au2firstlast[author].items():
                    for pmid in pmids:
                        pmid2firstlast[pmid] = mark
        return pmid2firstlast

    def prt_authors(self, coauthor2mark, prt_top_n=15):
        """Print authors w/most collaborations and relevant researchers"""
        for idx, (author, cnt) in enumerate(self.author2cnt.most_common()):
            seen = self._get_mark(author, coauthor2mark)
            firstlast = self.get_firstlast(author)
            if idx < prt_top_n or seen != '':
            ## if True:
                print('{S:>6} {N:3} {FL:6} {AU}'.format(FL=firstlast, S=seen, N=cnt, AU=author))
        print('{A} authors (appeared N times, M people): {LST}'.format(
            A=len(self.author2cnt),
            LST=sorted(Counter(self.author2cnt.values()).items())))

    def get_firstlast(self, author):
        """Get a string describing First and Last author affiliation"""
        if author not in self.au2firstlast:
            return ''
        firstlast = self.au2firstlast[author]
        # First author counts
        ret = 'F{N}'.format(N=len(firstlast['F'])) if 'F' in firstlast else '..'
        # Second author counts
        ret += 'f{N}'.format(N=len(firstlast['f'])) if 'f' in firstlast else '..'
        # Last author counts
        ret += 'L{N}'.format(N=len(firstlast['L'])) if 'L' in firstlast else '..'
        return ret

    def _init_au2firstlast(self, authors):
        """Get the number of times an author is in the first, middle, or last position"""
        au2firstlast = defaultdict(lambda: defaultdict(set))  # Author -to- First, Middle, Last
        for pmid, paper in self.pmid2paper.items():
            icite = paper.pmid2icite[pmid]
            authors = icite.dct['authors']
            ## print('PPPPPPPPPPPP', pmid, paper)
            if authors:
                ## print('PPPPPPPPPPPP', authors)
                # Count instances of being first or last author
                au2firstlast[authors[0]]['F'].add(pmid)
                au2firstlast[authors[-1]]['L'].add(pmid)
                if len(authors) > 2:
                    au2firstlast[authors[1]]['f'].add(pmid)
        return au2firstlast

    def _init_au2cnt(self):
        """List each co-author w/the number of times they co-author in pmid2paper"""
        authors = Counter()
        for pmid, paper in self.pmid2paper.items():
            icite = paper.pmid2icite[pmid]
            for author in icite.dct['authors']:
                authors[author] += 1
            ## print('sssssssssss', pmid, icite.dct)
        return authors

    @staticmethod
    def _get_mark(author, coauthor2mark):
        """Get a match, given a full author name"""
        for coau, mrk in coauthor2mark.items():
            if coau in author:
                return mrk
        return ''


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



"""Given a PubMed ID (PMID), download a list of publications which cite and reference it"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from pmidcite.icite.dnldr.pmid_dnlder_base import NIHiCiteDownloaderBase
from pmidcite.icite.entry import NIHiCiteEntry


class NIHiCiteDownloaderOnly(NIHiCiteDownloaderBase):
    """Given a PubMed ID (PMID), download a list of publications which cite and reference it"""

    ##def __init__(self, details_cites_refs=None, nih_grouper=None):
    ##    super(NIHiCiteDownloaderOnly, self).__init__(details_cites_refs, nih_grouper)

    def get_icites(self, pmids):
        """Download NIH iCite data for requested PMIDs"""
        pmid2nihentry = {o.pmid: o for o in self._dnld_icites(pmids)}
        return [pmid2nihentry[pmid] for pmid in pmids if pmid in pmid2nihentry]

    def _dnld_icites(self, pmids):
        """Download a list of NIH citation data for PMIDs"""
        nihdicts = self.api.dnld_nihdicts(pmids)
        if nihdicts:
            s_get_group = self.nihgrouper.get_group
            return [NIHiCiteEntry(d, s_get_group(d['nih_percentile'])) for d in nihdicts]
        return []

    def get_icite(self, pmid):
        """Load or download NIH iCite data for requested PMID"""
        nih_dict = self.api.dnld_nihdict(pmid)
        if nih_dict:
            return NIHiCiteEntry(
                nih_dict,
                self.nihgrouper.get_group(nih_dict['nih_percentile']))
        return None


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Read a file created by pmidcite and write simple text file of PMIDs"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from pmidcite.icite.entry import NIHiCiteEntry


def get_details_cites_refs(verbose, load_citations, load_references, no_references):
    """Optionally download and print citations and/or references of a specified papers"""
    # If verbose, download and print all citations and references
    if verbose:
        details_cites_refs = set(NIHiCiteEntry.associated_pmid_keys)
        # TBD: obsolete no_references
        if no_references:
            details_cites_refs = details_cites_refs.difference(NIHiCiteEntry.refkey)
        return details_cites_refs
    details_cites_refs = set()
    if load_citations:
        details_cites_refs.update(NIHiCiteEntry.citekeys)
    if load_references:
        details_cites_refs.update(NIHiCiteEntry.refkey)
    # TBD: obsolete no_references
    if no_references:
        details_cites_refs = details_cites_refs.difference(NIHiCiteEntry.refkey)
    return details_cites_refs


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.
